package java.sql;

/* 
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2000 All Rights Reserved
 */
import java.util.Calendar;
 
public class Timestamp extends java.util.Date {

   private transient int nanos;

   

   public Timestamp(long time)
   {
      super((time/1000)*1000);
     
      if (time < 0)
      {
         setTime(((time/1000-1)*1000));
         nanos = nanos + 1000000;
      }
      else
         nanos = (int) ((time%1000)*1000000); 
   }

   public static Timestamp valueOf(String s)
   {
      int posDelimSp, posDelimPd, posDelim1, posDelim2, count = 0;
      int year, month, day;
      int hour, minute, second;
      int nano = 0;
      String dateString, timeString, nanoString = null;
      String errMsg = "Timestamp must be in the form:  yyyy-mm-dd-hh.mm.ss.nnnnnn";

      if (s == null || s.length() > "yyyy-mm-dd-hh.mm.ss.nnnnnn".length())
         throw new java.lang.IllegalArgumentException(errMsg);

      posDelimSp = s.indexOf('-',0);
      if (posDelimSp > 0) {
      	count++;
      }
	while (posDelimSp > 0 && count >= 1 && count < 3) {
		posDelimSp = s.indexOf('-',posDelimSp + 1);
		count++;
	} 
	
      if (posDelimSp > 0)
      {
         dateString = s.substring(0, posDelimSp);
         timeString = s.substring(posDelimSp+1);
      }
      else
         throw new java.lang.IllegalArgumentException(errMsg);

      if (dateString == null || (dateString.length() > "yyyy-mm-dd".length()))
         throw new java.lang.IllegalArgumentException(errMsg);

      posDelim1 = dateString.indexOf('-',0);
      posDelim2 = dateString.lastIndexOf('-');

      if ( (posDelim1 > 0) && (posDelim2 != dateString.length()-1) && ((posDelim2-posDelim1) > 1) )
      {
         year  = Integer.parseInt(dateString.substring(0, posDelim1)) - 1900;
         month = Integer.parseInt(dateString.substring(posDelim1+1, posDelim2)) - 1;
         day   = Integer.parseInt(dateString.substring(posDelim2+1));
      }
      else 
         throw new java.lang.IllegalArgumentException(errMsg);
      
      if (timeString == null)
         throw new java.lang.IllegalArgumentException(errMsg);

	count = 0;
      posDelimPd = timeString.indexOf('.',0);
	if (posDelimPd > 0) {
		count++;
	}
	while (posDelimPd > 0 && count >= 1 && count < 3) {
		posDelimPd = timeString.indexOf('.',posDelimPd + 1);
		count++;
	}

      if (posDelimPd > 0)
      {
         nanoString = timeString.substring(posDelimPd+1);
         timeString = timeString.substring(0,posDelimPd);
      }

      if (timeString == null || (timeString.length() > "hh.mm.ss".length()))
         throw new java.lang.IllegalArgumentException(errMsg);

      posDelim1 = timeString.indexOf('.',0);
      posDelim2 = timeString.lastIndexOf('.');
      
      if ( (posDelim1 > 0) && (posDelim2 != s.length()-1) && ((posDelim2-posDelim1) > 1) )
      {
         hour   = Integer.parseInt(timeString.substring(0, posDelim1));
         minute = Integer.parseInt(timeString.substring(posDelim1+1, posDelim2));
         second = Integer.parseInt(timeString.substring(posDelim2+1));
      }
      else 
         throw new java.lang.IllegalArgumentException(errMsg);

      if (nanoString != null)
      {
         int nanoLen = nanoString.length();

         if (nanoLen < 6)
            nanoString = nanoString + "000000".substring(0, 6-nanoLen);

         nano = Integer.parseInt(nanoString);
      }

	Calendar cal = Calendar.getInstance();
	cal.set(Calendar.YEAR, year);
	cal.set(Calendar.MONTH, month);
	cal.set(Calendar.DAY_OF_MONTH, day);
	cal.set(Calendar.HOUR, hour);
	cal.set(Calendar.MINUTE, minute);
	cal.set(Calendar.SECOND, second);
	Timestamp ts = new Timestamp(cal.getTime().getTime());
	ts.setNanos(nano);
	return ts;
   }

   public String toString()
   {
   	Calendar cal = Calendar.getInstance();
   	cal.setTime(new Date(this.getTime()));
   	int year         = cal.get(Calendar.YEAR);
   	int month        = cal.get(Calendar.MONTH);
      int day          = cal.get(Calendar.DAY_OF_MONTH);
      int hour         = cal.get(Calendar.HOUR);
      int minute       = cal.get(Calendar.MINUTE);
      int second       = cal.get(Calendar.SECOND);
      String theYear   = Integer.toString(year);
      String theMonth  = (month >= 10) ? (Integer.toString(month)): ("0"+month);
      String theDay    = (day >= 10) ? (Integer.toString(day)): ("0"+day);
      String theHour   = (hour >= 10) ? (Integer.toString(hour)): ("0"+hour);
      String theMinute = (minute >= 10) ? (Integer.toString(minute)): ("0"+minute);
      String theSecond = (second >= 10) ? (Integer.toString(second)): ("0"+second);
      String theNano   = Integer.toString(nanos);

      // format nano
      if (nanos != 0)
      {
         int  nanoLen = theNano.length(), i = 0;
         
         theNano = "000000".substring(0, 6-nanoLen) + theNano;
    /*
         char nanoArr[] = new char[nanoLen];
         theNano.getChars(0,nanoLen,nanoArr,0);

         while (i < nanoArr.length)
         {
            if (nanoArr[nanoArr.length-i-1] != '0')
               break;
            i++;
         }
            
         if (i > 0)
            theNano = theNano.substring(0, 6-i);
     */
      }

      return (theYear + "-" + theMonth + "-" + theDay + "-" +
              theHour + "." + theMinute + "." + theSecond + "." +
              theNano);
   }

   public int getNanos()
   {
      return nanos;
   }

   public void setNanos(int n)
   {
      if (n < 0 || n > 999999)
         throw new java.lang.IllegalArgumentException("Invalid nano value");

      nanos = n;
   }

   public boolean equals(Timestamp ts)
   {
      return (super.equals(ts) && nanos == ts.nanos);
   }

   public boolean equals(Object ts)
   {
      if (ts instanceof Timestamp)
         return (this.equals((Timestamp)ts));

      return false;
   }

   public boolean before(Timestamp ts)
   {
      long lTime = this.getTime();
      long rTime = ts.getTime();

      if (lTime < rTime)
         return true;
      else if (lTime > rTime)
         return false;
      else
      {
         if (nanos < ts.nanos)
            return true;

         return false;
      }
   }

   public boolean after(Timestamp ts)
   {
      long lTime = this.getTime();
      long rTime = ts.getTime();

      if (lTime > rTime)
         return true;
      else if (lTime < rTime)
         return false;
      else
      {
         if (nanos > ts.nanos)
            return true;
 
         return false;
      }
   }

}

