package java.sql;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 1998, 2000 All Rights Reserved
 */

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class DriverManager {

   private static Vector drivers        = new Vector();
   private static int loginTimeout      = 0;
   private static java.io.OutputStream logWriter 		    = null;

   private static final String ERR_CONN = "08001";

   static
   {
      // load drivers from local file system
	// preloadDrivers(); 
   }

   private static void preloadDrivers()
   {
      String driverList = System.getProperty("jdbc.drivers");
      int posDelim = (driverList != null)? (driverList.indexOf(':',0)): -1;

//	println ("JDBC DriverManager:  jdbc.drivers" + driverList);

      while (posDelim > 0)
      {
         String driver = driverList.substring(0, posDelim);

         try
         {
//            println ("JDBC DriverManager:  loading " + driver);
            Class.forName(driver); //, true, ClassLoader.getSystemClassLoader());
         }
         catch (ClassNotFoundException e)
         {
//            println ("JDBC DriverManager:  fail loading " + driver);
         }

         driverList = driverList.substring(posDelim+1);
         posDelim = driverList.indexOf(':',0);
      }
   }

   private DriverManager() {};

   

   public static synchronized Connection getConnection(String url, String user, String password) throws SQLException
   {
      Hashtable info = new Hashtable();
      info.put("user", user);
      info.put("password", password);

      return getConnection(url, info); 
   }

   public static synchronized Connection getConnection(String url) throws SQLException
   {
      return getConnection(url, new Hashtable()); 
   }

    private static synchronized Connection getConnection (String url,
                                                         Hashtable info) throws SQLException

   {
      int size = drivers.size();
      SQLException rex = null;

      if (url != null) 
      {
         for (int i = 0; i < size; i++)
         {
            Driver driver = (Driver)drivers.elementAt(i);

            try
            {
               if (driver.acceptsURL(url))
               {
                  Connection c = driver.connect(url, info);
                  if (c != null)
                  {
                     return c;
                  }
               }
            }
            catch (SQLException e)
            {
               if (rex == null) 
                  rex = e;
            }
         }
      }

      if (rex != null)
         throw rex;

//      println("JDBC DriverManager:  Driver not found: " + url);

      throw new SQLException("Driver not found: " + url, ERR_CONN );
   }

   public static Driver getDriver(String url) throws SQLException
   {
//      ClassLoader invokerCL = com.ibm.oti.vm.VM.callerClassLoader();
      int size = drivers.size();

      if (url != null)
      {
         for (int i = 0; i < size; i++)
         {
            Driver driver = (Driver)drivers.elementAt(i);
//            if (!checkCallerClass(invokerCL, driver))
//               continue;

            try
            {
               if (driver.acceptsURL(url))
                  return driver;
            }
            catch (SQLException e)
            {
            }
         }
      }

//      println("JDBC DriverManager:  No suitable driver for " + url);

      throw new SQLException ("JDBC DriverManager:  No suitable driver for " + url);
   }

   
   public static synchronized void registerDriver(Driver driver) throws SQLException
   {
      drivers.addElement(driver);
   }

   public static synchronized void deregisterDriver(Driver driver) throws SQLException
   {
//      ClassLoader invokerCL = com.ibm.oti.vm.VM.callerClassLoader();
      int size = drivers.size();
      int i;

      for (i = 0; i < size; i++)
      {
         Driver d = (Driver)drivers.elementAt(i);
         if (d == driver)
         {
//             if (!checkCallerClass(invokerCL, driver))
//               throw new SecurityException("Not authorized to deregister");
            break;
         }
      }

      //  Safely remove driver
      if (i < size)
         drivers.removeElementAt(i);
   }

   public static Vector getDrivers()
   {
//      ClassLoader invokerCL = com.ibm.oti.vm.VM.callerClassLoader();
      Vector validDrivers = new Vector();
      int size = drivers.size();

      for (int i = 0; i < size; i++)
      {
         Driver driver = (Driver)drivers.elementAt(i);
//         if (!checkCallerClass(invokerCL, driver))
//            continue;

         validDrivers.addElement(driver);
      }

      return validDrivers;
   }

   public static void setLoginTimeout(int seconds)
   {
      loginTimeout = seconds;
   }

   public static int getLoginTimeout()
   {
      return loginTimeout;
   }




}
