/******************************************************************************/
/*   IBM Confidential                                                         */
/*                                                                            */
/*   OCO Source Materials.                                                    */
/*                                                                            */
/*   5622-388 (C) Copyright IBM Corp. 1999.                                   */
/*                                                                            */
/*   The source code for this program is not published or otherwise           */
/*   divested of its trade secrets, irrespective of what has been             */
/*   deposited with the U.S. Copyright Office.                                */
/*                                                                            */
/*   Product ID: 5648C61                                                      */
/*   2001/01/17         Created for Palm a CLP interface using common         */
/*                      code base. Prototype is based on the DB2e utilities   */
/*                      application but has simpler functionality             */
/******************************************************************************/
/* DB2e CLP
 */


#if PROFILE
#include "HostControlProfile.h"
#endif

#include "CLPpalmRsc.h"
#include "SampleCLP.h"
#include "SampleCLPmsg.h"

//To check memory leak, define MEM_CHECK, and define DEBUG_MEM_TRACE in oss.h.  051399jhc 
//#define MEM_CHECK
#ifdef MEM_CHECK
extern void oMemCheck();
#endif


/***********************************************************************
 * Structures for this module
 **********************************************************************/

/***********************************************************************
 * Global variables for this module
 **********************************************************************/
 CLPState    *clpstate = 0;
/***********************************************************************
 * Prototypes for internal functions
 **********************************************************************/

static int     StartApplication(void);
static int     StopApplication(void);
static void    EventLoop(void);
static Boolean MenuEventHandler(EventPtr event);
static Boolean DB2MainHandler(EventPtr event);
static Boolean CLPHandler(EventPtr event);
static void    CLPNewStatement();
static void    CLPExecute();
void           CLPStdout(char *s);
static void    CLPUpdateScrollBar(Word fldID, Word barID);
static void    CLPPageScroll(WinDirectionType direction, Word fldID, Word barID);
static void    CLPScroll(Short numLines, Word fldID, Word barID);
static void DB2MainDrawForm(void);
static void    CLPDrawForm(void);
static int       DB2Execute(CharPtr query);

#define ErrAlert0(s1)    FrmCustomAlert(ErrorAlert, s1, " ", " ")
#define InfoAlert0(s1)   FrmCustomAlert(WarningAlert, s1, " ", " ")

/*--------------------------------*/
// general routines
static VoidPtr GetObjectPtr(Word objectID)
{
  FormPtr frmP = NULL;
  frmP = FrmGetActiveForm();
  return (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID)));
}
/* ****************************************************************************
   *****                                                                  *****
   *****       Pilot Main                                                 *****
   *****                                                                  *****
   **************************************************************************** */

DWord  PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
  int rc = 0;
  char chow[3]="lg";
  switch (cmd) 
	{
	case sysAppLaunchCmdNormalLaunch:
	  {
		rc = StartApplication();
		if (rc) {
		  ErrAlert0(chow); }
		if (rc) goto exit;
		EventLoop();
		StopApplication();
	  }
	  break;
	case sysAppLaunchCmdFind:   
	  {
		break;
	  }
	default: 
	  break;
	}
 exit:
  return rc;
}

static int StartApplication(void)
{
  
#if PROFILE  
  ProfileInit(100,10);
  ProfileStart();
#endif
 
  FrmGotoForm(DB2MainForm);
  
}

static int StopApplication(void)
{
  int rc = 0;
  FrmCloseAllForms();
   if (clpstate)
	{
      CLPStopEngine(clpstate);
      free(clpstate);
	}
  if (rc) goto exit;
 exit:
#if PROFILE
  ProfileStop();
  ProfileDump("DUMP");
#endif
  
#ifdef MEM_CHECK
  oMemCheck();
#endif
  return rc;
}

/* ****************************************************************************
   *****                                                                  *****  
   *****       Event Loop                                                 *****
   *****                                                                  *****
   **************************************************************************** */

static void EventLoop(void)
{
  unsigned short err = 0;
  int formID = 0;
  FormPtr form = NULL;
  EventType event;
  
  
  do
  {
    EvtGetEvent(&event, -1);
	
    if (SysHandleEvent(&event)) continue;
    if (MenuHandleEvent((MenuBarType *)0, &event, &err)) continue;
	
    if (event.eType == frmLoadEvent)
	  {
		formID = event.data.frmLoad.formID;
		form = FrmInitForm(formID);
		FrmSetActiveForm(form);
		switch (formID)
		  {
		  case DB2MainForm:
			FrmSetEventHandler(form, DB2MainHandler);
			break;    
		
		  case CLPForm:
			FrmSetEventHandler(form, CLPHandler);
			break;    
		 
		  }
	  }
    FrmDispatchEvent(&event);
  } while(event.eType != appStopEvent);
}

/* ----------------------------------------------------*/
/*             DB2 Main form                           */
/* ----------------------------------------------------*/
static Boolean DB2MainHandler (EventPtr event)
{
  int handled = 0;
  int rc = 0;
  int n = 0;
  

  switch (event->eType) 
	{
	case frmOpenEvent:
	  DB2MainDrawForm();
	  break;
	  
	case ctlSelectEvent:  // A control button was pressed and released
	  switch (event->data.ctlEnter.controlID) 
        {
		
		case DB2MainCLPButton:
		       
		  FrmGotoForm(CLPForm);
		  handled = true;
		  break;
		
        }
	}
  return handled;
}

static void DB2MainDrawForm(void)
{ 
  FormPtr         frm = NULL;
 
  frm = FrmGetActiveForm();
  FrmDrawForm(frm);
}




/* ------------------------------------------------ */
/*             CLP         form                     */
/* ------------------------------------------------ */
static Boolean CLPHandler (EventPtr event)
{
   int handled = 0;
   int objIndex;
   FormPtr frm;
   Word objID;
   int rc;
   
   switch (event->eType)
   {
   case frmOpenEvent:
        CLPDrawForm();
        handled = 1;
        
           clpstate = (CLPState *) malloc(sizeof(CLPState));
        if (!clpstate)
	      {
	        CLPStdout(_TEXT(CLP_INTERNAL_ERROR2));
	      }
        clpstate->mode  = CLP_INTERACT;
        clpstate->fpIN  = 0;
        clpstate->fpOUT = 0;

        // Start DB2 Everyplace engine
        rc = CLPStartEngine(clpstate);
  
        if (rc)
	      {
	       CLPStdout ( _TEXT(CLP_DB2_NOSTART) );
	       if (clpstate)
		     free(clpstate);
		     clpstate = 0;
	       }
	       
        break;
        
   case keyDownEvent:
        if (event->data.keyDown.chr == pageUpChr) {
            frm = FrmGetActiveForm ();
			objIndex = FrmGetFocus(frm);
			if (objIndex >=0 ) {
			  objID = FrmGetObjectId(frm, objIndex);
			  if (objID == CLPStatementScrollBar) {    
                 CLPPageScroll(winUp, CLPStatementField, CLPStatementScrollBar);
              } else {
                 CLPPageScroll(winUp, CLPResultField, CLPResultScrollBar);
              }
              handled = true;
            }
        } else if (event->data.keyDown.chr == pageDownChr) {
            frm = FrmGetActiveForm ();
			objIndex = FrmGetFocus(frm);
			if (objIndex >=0 ) {
			   objID = FrmGetObjectId(frm, objIndex);
			   if (objID == CLPStatementScrollBar) {    
                   CLPPageScroll(winDown, CLPStatementField, CLPStatementScrollBar); 
               } else {
                   CLPPageScroll(winDown, CLPResultField, CLPResultScrollBar);
               }
               handled = true;
            }
        } else { // An ordinary ASCII character was entered.  Have the form
				// give the field the character.  Then check to see if the 
				// scrolling changed.
				frm = FrmGetActiveForm ();
				FrmHandleEvent (frm, event);
				objIndex = FrmGetFocus(frm);
				if (objIndex >=0) {
			       objID = FrmGetObjectId(frm, objIndex);
			       if (objID == CLPStatementField) {    
                      CLPUpdateScrollBar(CLPStatementField, CLPStatementScrollBar); 
                   } else {
                      CLPUpdateScrollBar(CLPResultField, CLPResultScrollBar);
                   }
            	   
				   handled = true;
				}
		}
		break;   
   case sclRepeatEvent:
   
		objID = event->data.sclRepeat.scrollBarID;
      switch (objID)
      {
      	case CLPStatementScrollBar:
	        CLPScroll(event->data.sclRepeat.newValue - event->data.sclRepeat.value,
	        	CLPStatementField, CLPStatementScrollBar); 
	        break;
	        
	      case CLPResultScrollBar:
          CLPScroll(event->data.sclRepeat.newValue - event->data.sclRepeat.value,
            CLPResultField, CLPResultScrollBar);
          break;
        
        default:
        	break;
      }
     handled = 1; 
		break;
				
   case ctlSelectEvent: // A control button was pressed and released
      
        switch (event->data.ctlEnter.controlID)
        {
         
             
        case CLPInfoButton: 
             if(!clpstate)
             exit(-1);
						#if DEBUG_VERSION
								  FrmCustomAlert(InfoAlert, CLP_VERSION, " ", " ");
						#else
								  FrmCustomAlert(InfoAlert, CLP_VERSION, " ", " ");
						#endif
								  handled = true;
        case CLPNewButton:
              if(!clpstate)
              exit(-1);
              else
              CLPNewStatement();
              handled = true;
              break;
        case CLPExecuteButton:
             if(!clpstate)
             exit(-1);
             else
              {
               frm = FrmGetActiveForm ();
               CLPExecute(clpstate);
               }
              handled = true;
              break;
        /* case CLPExitButton:
              
              //MemSet(event, sizeof(EventType), 0);
		          event->eType = appStopEvent;
		          EvtAddEventToQueue(event);
		          handled = true;
		          
              break;
         */     
        }
   }
      
   return handled;
   
}


static void CLPUpdateScrollBar (Word fldID, Word barID)
{
	Word scrollPos;
	Word textHeight;
	Word fieldHeight;
	Short maxValue;
	FieldPtr fld;
	ScrollBarPtr bar;
    FormPtr frm;
    
    frm = FrmGetActiveForm();
 
    bar = (ScrollBarPtr) GetObjectPtr(barID);
	fld = (FieldPtr) GetObjectPtr(fldID);
	                                          
	// get the values necessary to update the scroll bar.
	FldGetScrollValues (fld, &scrollPos, &textHeight,  &fieldHeight);

	if (textHeight > fieldHeight)
		maxValue = textHeight - fieldHeight;
	else if (scrollPos)
		maxValue = scrollPos;
	else
		maxValue = 0;

	SclSetScrollBar (bar, scrollPos, 0, maxValue, fieldHeight-1);
  
}

static void CLPPageScroll (WinDirectionType direction, Word fldID, Word barID)
{
	Short value;
	Short min;
	Short max;
	Short pageSize;
	Word linesToScroll;
	FieldPtr fld;
	ScrollBarPtr bar;
    FormPtr frm;
    
    frm = FrmGetActiveForm();
 
	fld = (FieldPtr) GetObjectPtr(fldID);
	if (FldScrollable (fld, direction))
		{
		linesToScroll = FldGetVisibleLines (fld) - 1;
		FldScrollField (fld, linesToScroll, direction);

		// Update the scroll bar.
		bar = (ScrollBarPtr) GetObjectPtr (barID);
		SclGetScrollBar (bar, &value, &min, &max, &pageSize);

		if (direction == winUp)
			value -= linesToScroll;
		else
			value += linesToScroll;
		
		SclSetScrollBar (bar, value, min, max, pageSize);
		}
}

static void CLPScroll (Short linesToScroll, Word fldID, Word barID)
{
	Word				blankLines;
	Short				min;
	Short				max;
	Short				value;
	Short				pageSize;
	FieldPtr			fld;
	ScrollBarPtr	bar;
	FormPtr frm;
	
	frm = FrmGetActiveForm();
	
	fld = (FieldPtr) GetObjectPtr (fldID);
  if (linesToScroll < 0)
		{
		blankLines = FldGetNumberOfBlankLines (fld);
		FldScrollField (fld, -linesToScroll, winUp);
		
		// If there were blank lines visible at the end of the field
		// then we need to update the scroll bar.
		if (blankLines)
			{
			// Update the scroll bar.
			bar = (ScrollBarPtr) GetObjectPtr (barID);
			SclGetScrollBar (bar, &value, &min, &max, &pageSize);
			if (blankLines > -linesToScroll)
				max += linesToScroll;
			else
				max -= blankLines;
			SclSetScrollBar (bar, value, min, max, pageSize);
			}
		}

	else if (linesToScroll > 0)
		FldScrollField (fld, linesToScroll, winDown);
 FrmDrawForm(FrmGetActiveForm());
}

static void CLPDrawForm(void)
{
  FormPtr frm;
  frm = FrmGetActiveForm();
  FrmDrawForm(frm);
}

static void CLPNewStatement() {
  FormPtr frm;
  FieldPtr stField;
  FieldPtr resField;
  
  
  frm = FrmGetActiveForm();
 
  stField = (FieldPtr) GetObjectPtr(CLPStatementField);
  resField = (FieldPtr) GetObjectPtr(CLPResultField);
  FldFreeMemory(stField);
  FldEraseField(stField);
  FldFreeMemory(resField);
  FldEraseField(resField);
  CLPUpdateScrollBar(CLPStatementField, CLPStatementScrollBar);
  CLPUpdateScrollBar(CLPResultField, CLPResultScrollBar);
  FrmDrawForm(frm);  
}

int QueryInsPos =0;
int ResultInsPos = 0;

static void CLPExecute() {
   FormPtr frm;
  FieldPtr stField;
  FieldPtr resField;
  CharPtr fieldText;

  CharPtr resText = NULL;
  int newQueryInsPos;
  int rc = 0;
  
  frm = FrmGetActiveForm();
  stField = GetObjectPtr(CLPStatementField);
  newQueryInsPos = FldGetInsPtPosition(stField);
  
  fieldText = FldGetTextPtr(stField) + QueryInsPos;
 
  if (fieldText) {
      resField = GetObjectPtr(CLPResultField);
      rc = DB2Execute(fieldText);
      if (rc) goto exit;      
  }   
 
  FrmDrawForm(frm);
exit:
  return;    
}

void CLPStdout(char *s) {
   int len = StrLen(s); 
   FieldPtr resField;
   FormPtr frm;
   frm = FrmGetActiveForm();

   resField = GetObjectPtr(CLPResultField);
   FldGrabFocus(resField);
   FldInsert(resField, s, len);
   ResultInsPos += len;
   if (ResultInsPos > FldGetMaxChars(resField)) {  
     FldFreeMemory(resField);
     FldEraseField(resField);
     ResultInsPos = 0;
   }
   FldSetInsPtPosition(resField, ResultInsPos);

   CLPUpdateScrollBar(CLPResultField, CLPResultScrollBar);
   FldDrawField(resField);
}

//static int DB2Execute(CharPtr query)
static int DB2Execute(CharPtr query)
{
  
   int          clp_status = CLP_CONTINUE;
 
	 if(!clpstate)
	  exit(-1);
	 else
	 {
	    _tcscpy(clpstate->cmd, query);
	    clp_status = CLPcommand_interpreter(clpstate);
	    //CLPStdout(clpstate->err_string);
		if (CLP_QUIT == clp_status )
						goto exit;
	  }
   
     
   exit:
  return 0;	 
  
}

  
