///////////////////////////////////////////////////////////////
//
// Source File Name = SampleCLP.c
//
// (C) Copyright IBM Corp. 2000, 2001
// All Right Reserved.
// Licensed Materials - Property of IBM
//
// DB2 Everyplace Sample Application - CLP
// 
// This application demonstrates how to use CLI to
// write a simple command line processor that 
// accepts SQL statements and execute them.
// This application is portable to different platforms:
//    Palm, Neutrino, E. Linux, WinCE, and EPOC.
// It also includes import/export tools.
//
//
// History:
// d11182000ctl     fix a problem on reading integer w/ byte alignment
//  11282000ctl     add $chmod support and port to EPOC
//  02192001ctl     fix BLOB support, describe and $file commands
//  03192001ctl     add "connect to c:/temp/" support
//  03192001xh      integrate with Palm
//  03232001ctl     add timestamp support
//  03232001xh      timestamp support for Palm 
//  04022001ctl     add $table command
//  05112001thf     fix printing problem of DESCRIBE output
//  05232001jca     fix import problem for WinCE
//  06112001jca     fix $table command for WinCE
//  06122001jca     fix import problem with double quotes
//  06202001xh      port to EPOC5
//  06262001thf     fix broken DESCRIBE command output on PalmOS
//  06272001kmt     Change sprintf to _stprintf
//  07022001xh      remove UNDER_EP
//  07132001ctl     Win32 unicode mode
//  07232001ctl     check rc for SQLGetData
//  08022001kmt     EPOC Release 6 support
//  08272001kmt     nb_read = # of bytes read; fix o_Fread
//  08282001xh      Internal error trace
//  09052001qz      Regression test on WinCE
//  09112001qz      Fix for d4182qz which affect the GUI. dfixqz
//  09192001kmt     Fix possible memory errors.
//  09192001qz      enhance sampleCLP to support multiple statement handles d4254qz
//  09282001jgu     no extra \n for WinCE
//  10122001xh      Clean up code for supporting multiple stmt handles & transactions
//  10232001xh      Port new commands of multiple stmt handles to all platform
//  10242001jgu		Fix \n for regression test on CE
//  01072002kmt     Free memory
///////////////////////////////////////////////////////////////

#include "SampleCLP.h"
#include "SampleCLPmsg.h"

//#define ERR_TRACE 1 //d4109

// SQL Statements
static const SQLTCHAR SQL_SELECT[] = _TEXT("SELECT * FROM ");
static const SQLTCHAR SQL_INSERT[] = _TEXT("INSERT INTO ");
static const SQLTCHAR SQL_VALUES[] = _TEXT(" VALUES(");

// format
static const SQLTCHAR DB2e_intFormat[] = _TEXT("%d,");
static const SQLTCHAR DB2e_timeFormat[] = _TEXT("\"%d.%d.%d\",");
static const SQLTCHAR DB2e_timestampFormat[] = _TEXT("\"%d-%d-%d-%d.%d.%d.%d\","); // d03212001ctl

static const SQLTCHAR CRLF[] = _TEXT("\r\n");

// This is for PocketPC
#if defined(oWINCE)
HWND hwndLV;
#endif


#if defined(oWINCE)
OSS_Encoding=-1; //d3041
#endif


//for EPOC; d3649 xh 06202001
#if defined(oEPOC)
extern void callback_print1_ep(void* aepSupport, SQLTCHAR* msg);
#endif
/********************************************************************************
** callable_CLP()
** This function exposes CLP function for other applications
** (compared w/ main() in SampleCLPMain.c.
********************************************************************************/
/*#if !defined(oWINCE)
int callable_CLP(SQLHENV  henv,
                 SQLHDBC  hdbc,
                 SQLHSTMT hstmt,
                 SQLTCHAR *cmd)
#else
    int callable_CLP(SQLHENV  henv,
                     SQLHDBC  hdbc,
                     SQLHSTMT hstmt,
                     SQLTCHAR *cmd,
                     HWND hWnd)

#endif
{
    SQLRETURN      rc;
    CLPState*      clpstate;

    if (!(clpstate = malloc(sizeof(CLPState))))
      return -1;

    // Allocate and initialize CLPState structure
    clpstate->henv   = henv;
    clpstate->hdbc   = hdbc;
    clpstate->hstmt  = hstmt;
#if defined(oWINCE)
    clpstate->mode   = CLP_INTERACT;
#else
    clpstate->mode   = CLP_INTERACT_DEVICE;
#endif
    clpstate->fpIN   = 0;
    clpstate->fpOUT  = 0;

    _tcscpy(clpstate->cmd, cmd);

#if defined(oWINCE)
    hwndLV = hWnd;
#endif

    rc = CLPcommand_interpreter(clpstate);

    free(clpstate);

    return(rc);
}
*/


#if !defined(oWINCE) 
 int callable_CLP(CLPState *clpstate)
#else 
 int callable_CLP(CLPState *clpstate, HWND hWnd)
#endif
{
 SQLRETURN      rc;

    
#if defined(oWINCE)
    hwndLV = hWnd;
#endif
    
    rc = CLPcommand_interpreter(clpstate);

    return(rc);
}

//Palm CLP begin
#if defined (oPALMOS)
char oToUpper(char ch)
{
    if ((ch >= 'a')&&(ch <= 'z'))
        return( ch - 'a' + 'A');
    else
        return ch;    
}
#endif 
//Palm CLP end 



//#if (defined (oEPOC) && defined(UNICODE)) || defined(oPALMOS)
// implemented in SampleCLP.c for EPOC & Palm 
// it has some limitations:
// 1) the formatprec here is only the preceding of a real format string
// e.g. for format "$table %s %s", formatprec is "$table"
// 2) the maximum number of resulting strings is 3, if the integer to be
// returned is less than 3 then, res2 or/and res3 is NULL upon returning 
//#if (defined (oEPOC) && defined(UNICODE)) || defined(oPALMOS) 
int oSscanf(SQLTCHAR *buf, SQLTCHAR *formatprec, 
            SQLTCHAR *res1, SQLTCHAR *res2, SQLTCHAR *res3)
{ 
  SQLTCHAR  *pstr, *bstr;
  int numFields, len = 0;
  int i, j, num = 0;

  if(_tcsncmp(buf, formatprec, sizeof(formatprec))) 
    return -1;
    
  if(res1) memset(res1, 0, sizeof(res1));
  if(res2) memset(res2, 0, sizeof(res2));
  if(res3) memset(res3, 0, sizeof(res3));

  if(!res1)return -1; 
  else if(res2)
          if(res3) num = 3;
             else num = 2;
        else
           num = 1;

      //Parsing the input to get correct info
      pstr = buf + _tcslen(formatprec);
      while((*pstr == ' ')||(*pstr ==':')) pstr++; //bypassing space etc.

      numFields = 0;
      len = _tcslen((SQLTCHAR*)pstr);
      
      while ((len > 0)&&(numFields <= 3)) {
       if (*pstr != ' ') {
           numFields++;
           j = _tcslen((SQLTCHAR*)pstr);
           for (i = 0; i < j; i++) {
             if (numFields == 1) *(res1 + i) = *pstr;
             else if (numFields == 2) *(res2 + i) = *pstr;
             else if (numFields == 3) *(res3 + i) = *pstr;
             pstr++;
             len--;
            
             if ((*pstr == ' '||*pstr == '0') && (numFields == 1||(num == 3&&numFields == 2)) 
                 || len == 0)                   
                 {
                   if(numFields == 1) *(res1 + i + 1) = 0;
                   else if (numFields == 2) *(res2 + i + 1) = 0;
                   else if (numFields == 3) *(res3 + i + 1) = 0;
                    break;
                 }  
            }
           
         }
       else { 
             pstr++;
             len--;
         }
       }

  return numFields;
}
//#endif

/********************************************************************************
** DB2eDateToCSVDate()
** This function converts from YYYY-MM-DD to YYYYMMDD.
********************************************************************************/
static int DB2eDateToCSVDate(SQLTCHAR *s)
{
    s[4] = s[5];
    s[5] = s[6];
    s[6] = s[8];
    s[7] = s[9];
    return(8);                                // Returns the length of the data value
};

/********************************************************************************
** DB2eTimeToCSVTime()
** This function converts from HH:MM:SS to "HH.MM.SS"
********************************************************************************/
static int DB2eTimeToCSVTime(SQLTCHAR *s)
{
    s[9] = '"';
    s[8] = s[7];
    s[7] = s[6];
    s[6] = '.';
    s[5] = s[4];
    s[4] = s[3];
    s[3] = '.';
    s[2] = s[1];
    s[1] = s[0];
    s[0] = '"';
    return(10);                               // Returns the length of the data value
};

/********************************************************************************
** print1()
**   This function print out message (txt)
**   to the console or to a file if it was specified.
********************************************************************************/
void print1(CLPState *clpstate, SQLTCHAR *txt)
{
    SQLTCHAR   *newbuf = txt;
    int len, len1; // dfixqz add len1 and move definition here

#if defined(oWINCE)
    
    len1 = _tcslen(txt); // dfixqz change len to len1 to be different with len in the following block

	//d4486 jgu fix '\n' problem with regression test and SampleCLP on CE

	if (len1 > 0){
		if (txt[len1-1] == '\n')
		{
			txt[len1-1] = 0;
		}

		if (txt[0] == '\n'){
			AddItem(hwndLV, &txt[1], len1-1);
		}
		else{
			AddItem(hwndLV, txt, len1-1);
		}
     
	   if (txt[len1-1] == 0)  // dfixqz change txt[len1-1] back to '\n' for the following unicode conversion
		{
			txt[len1-1] = '\n';
		}
	}
#endif

/* #else d4182qz */
#if !defined(oEPOC) && !defined(oPALMOS)
    // 07132001ctl begin
#if defined(WIN32U) || defined(oWINCE) // add in oWINCE d4182qz 
    /* int dfixqz move definition to the above */     len = _tcslen(txt) * 3 + TEN_SPACES;;

    // character string in the buffer is in UCS-2 format, ie., 2 bytes/char
    // we need to convert to character using locale.
    newbuf = malloc(len);
    if (newbuf == 0) return;
    memset(newbuf, 0, len);
    WideCharToMultiByte(CP_ACP,               // Converts from TCHAR to CHAR; codepage
                        0, 
                        (LPTSTR) txt,
                        -1,                   // Null terminated string indicator
                        newbuf, 
                        len,                  // Buffer length
                        NULL,
                        NULL);
    len = strlen(newbuf);
    if (len <= 0) return;

#endif
    // 07132001ctl end
#if defined(REGRESSION)
    if (clpstate->regtest) { //d2786akl
        //09192001kmt _tcscpy(clpstate->err_string, newbuf); // 07132001ctl
        //09192001kmt CLPStdout((char*)clpstate->err_string);
        CLPStdout((char*)newbuf); //09192001kmt
    } else
#endif
    {
#if defined(WIN32U) //07132001ctl
        printf(newbuf);
        printf(_TEXT("\n"));
        free(newbuf);
#elif !defined(oWINCE)  //09212001kmt
        _tprintf(newbuf); 
        _tprintf(CRLF);
#endif
    }

#if defined(WIN32U) || defined(oWINCE) // 01072002kmt
    free(newbuf); newbuf = NULL;
#endif

#else
    //Palm CLP begin 
#if defined(oPALMOS)
    CLPStdout((char*)txt);
#if !defined(REGRESSION)
    CLPStdout((char*)"\n"); //07132001ctl
#endif
#endif
    //Palm CLP end 
 
    //EPOC CLP begin  d3649 xh06202001
#if defined(oEPOC)
  #ifdef REGRESSION //08022001kmt
    CLPStdout(txt);
  #else
    _tcscpy(clpstate->msg, txt);
    callback_print1_ep(clpstate->epSupport, txt);
  #endif // end of REGRESSION
#endif
    //EPOC CLP end
 
#endif    
//d4182qz #endif
}

void print2(CLPState *clpstate, SQLTCHAR *txt, SQLTCHAR *arg2)
{
#if defined(oWINCE)
    SQLTCHAR buf[BUFSIZE];  // should probably make it safer
    int len;

    memset(buf, 0, BUFSIZE*sizeof(SQLTCHAR));
    _stprintf(buf, txt, arg2);
    len = _tcslen(buf);
    if (buf[len-1] == '\n')
    {
        buf[len-1] = 0;
    }
    AddItem(hwndLV, buf, len);
#else
#if !defined(oEPOC) && !defined(oPALMOS)
    _tprintf(txt, arg2);
    _tprintf(CRLF); //07132001ctl
#else
    //for EPOC d3649 xh06202001
#if defined(oEPOC) 
    SQLTCHAR buf[BUFSIZE];
    memset(buf, 0, BUFSIZE*sizeof(SQLTCHAR));
    _stprintf(buf, (char*)txt, (char*)arg2);     
     
    _tcscpy(clpstate->msg, buf);
  #ifndef REGRESSION //08022001kmt
    callback_print1_ep(clpstate->epSupport, buf);
  #endif
#endif
   
    _stprintf((char*)clpstate->err_string, (char*)txt, (char*)arg2);
    //Palm CLP begin
#if defined(oPALMOS)
    CLPStdout((char*)clpstate->err_string);
    //CLPStdout((char*)"\n"); //d2786akl commented out
#endif
    //Palm CLP end  
#endif      
#endif
}




/********************************************************************************
** writeOut()
**   This function writes out message (txt)
**   to the console or to a file if it was specified.
********************************************************************************/
void writeOut(CLPState *clpstate, SQLTCHAR *txt)
{
    SQLTCHAR     *buf;    // 04022001ctl
    SQLTCHAR     *ptr;    // 04022001ctl
    long          length; // 04022001ctl
    SQLRETURN     rc;     // 04022001ctl
    long          i, min; // 04022001ctl
    char tmp[10]; //06112001jca
    char mbtmp[20]; //06112001jca
  
    if (txt)
    {
        if (clpstate->fpOUT)
        {
// 07132001ctl begin
            SQLTCHAR   *newbuf = txt;
#if defined(oPALMOS)
            o_Fwrite((char*)txt, sizeof(SQLTCHAR), _tcslen(txt), clpstate->fpOUT);
            o_Fwrite((char*)"\n", sizeof(SQLTCHAR), _tcslen("\n"), clpstate->fpOUT);    
#elif defined(oWINCE)
            o_Fwrite(txt, sizeof(SQLTCHAR), _tcslen(txt), clpstate->fpOUT);
            o_Fwrite(CRLF, sizeof(SQLTCHAR), 2, clpstate->fpOUT);
            o_Fflush(clpstate->fpOUT);
#elif defined(WIN32U)

            int     len = _tcslen(txt) * 3 + TEN_SPACES;;
            newbuf = malloc(len);
            if (newbuf == 0) return;
            memset(newbuf, 0, len);
            WideCharToMultiByte(CP_ACP,               // Converts from TCHAR to CHAR; codepage
                                0, 
                                (LPTSTR) txt,
                                -1,                   // Null terminated string indicator
                                newbuf, 
                                len,                  // Buffer length
                                NULL,
                                NULL);
            len = strlen(newbuf);
            if (len <= 0) return;
            o_Fwrite(newbuf, 1, len, clpstate->fpOUT);
            o_Fwrite("\n", 1, 1, clpstate->fpOUT);
            o_Fflush(clpstate->fpOUT);
            free(newbuf);
#else
            o_Fwrite(newbuf, sizeof(SQLTCHAR), _tcslen(txt), clpstate->fpOUT);
            o_Fwrite(CRLF, sizeof(SQLTCHAR), _tcslen(CRLF), clpstate->fpOUT); // 07132001ctl
            o_Fflush(clpstate->fpOUT);

#endif
// 07132001ctl end
        }
        else
        {
            print1(clpstate, txt);
        }

        // 04022001ctl begin
        if (TSTBIT(clpstate->mode, CLP_TABLE_MODE))
        { // hope that there are not many single quotes
            //length = (_tcslen(txt) + 50) * sizeof(SQLTCHAR);
            length = (_tcslen(txt) + 60) * sizeof(SQLTCHAR);//jc
            buf = (SQLTCHAR *) malloc(length);
            if (buf == 0)
            {
                print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_INTERNAL_ERROR2) );
                return;
            }

            ptr = buf;
            _tcscpy(ptr, _TEXT("INSERT INTO "));
            _tcscat(ptr, clpstate->outTable);
            _tcscat(ptr, _TEXT(" VALUES( "));
            ptr += _tcslen(ptr);
#if defined(oPALMOS)
            StrIToA((char*)ptr,clpstate->outTablecount);
#else
#if defined(oWINCE)
            _stprintf(tmp, "%d", clpstate->outTablecount); //06112001jca //06272001kmt
            MultiByteToWideChar(CP_ACP, 0, tmp, -1, mbtmp, 20);   //06112001jca     
            _tcscat(ptr, mbtmp);//06112001jca
#else
            _stprintf(ptr, "%d", clpstate->outTablecount);
#endif
#endif
            clpstate->outTablecount++;
            _tcscat(ptr, _TEXT(" , '"));
            ptr += _tcslen(ptr);

            min = _tcslen(txt);
            if (min > clpstate->outTableTextLen)
            {
                min = clpstate->outTableTextLen;
            }
            for (i = 0; i < min; i++) {
                if (txt[i] == '\'') {
                    *ptr = '\'';
                    ptr++;
                }
                *ptr = txt[i];
                ptr++;
            }

            *ptr = '\0';
            _tcscpy(ptr, _TEXT("')"));
                  
            rc = SQLExecDirect(clpstate->outhstmt, buf, SQL_NTS);
            if (rc)
            {
                print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_INTERNAL_ERROR) );
                return;
            }
                        
            memset(buf, 0, length);

            free(buf);
        }
        // 04022001ctl end

    }
}

/********************************************************************************
** ErrorReport()
**   This function reports the error code by using
**   SQLGetDiagRec().
********************************************************************************/
void ErrorReport(CLPState *clpstate)
{
    SQLRETURN     rc;
    SQLTCHAR          buffer[SQL_MAX_MESSAGE_LENGTH + 1] = {0};
    SQLTCHAR          sqlstate[SQL_SQLSTATE_SIZE + 1] = {0};
    SQLTCHAR      err_string[BUFSIZE],err_msg[BUFSIZE] = {0};
    SQLSMALLINT   length;
    long          native_err;

    if(clpstate->flag_hstmt)
    rc = SQLGetDiagRec(SQL_HANDLE_STMT, clpstate->hstmt, 1, sqlstate,
                       &native_err, buffer, SQL_MAX_MESSAGE_LENGTH + 1, &length);
    else
    rc = SQLGetDiagRec(SQL_HANDLE_STMT, clpstate->mhstmt[clpstate->current_mhstmt], 1, sqlstate,
                       &native_err, buffer, SQL_MAX_MESSAGE_LENGTH + 1, &length);

    if (rc == SQL_SUCCESS)
    { 
        // DB2e doesn't return errormsg!
        _stprintf((char*)err_string, (char*)_TEXT(CLP_MSG_STMT_ERROR), (char*)sqlstate);
        _tcscpy(clpstate->sqlstate, sqlstate);

       
         //d2551
        if((buffer[0] != '\0')&&(_tcscmp(sqlstate,_TEXT("42601")) == 0))
          {
           _stprintf((char*)err_msg, (char*)_TEXT(CLP_ERROR_TOKEN), (char*)buffer);
           _tcscpy(clpstate->err_string, buffer);
          }
 
        if (TSTBIT(clpstate->mode,CLP_EXPORT))
        {
            print1(clpstate, err_string);
            if((buffer[0] != '\0')&&(_tcscmp(sqlstate,_TEXT("42601")) == 0))                
                print1(clpstate, err_msg);
            
            #if 0 //defined(ERR_TRACE) //d4109
                print1(clpstate, buffer);
            #endif           
        }
        else
        {
            writeOut(clpstate, err_string);
            if((buffer[0] != '\0')&&(_tcscmp(sqlstate,_TEXT("42601")) == 0))
              writeOut(clpstate, err_msg);
            
            #if 0 //defined(ERR_TRACE) //d4109
              writeOut(clpstate, buffer);
            #endif
        }
    }
    else
        writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_ERROR_GET_STATE)); 
}

/********************************************************************************
** CLPSetStmtMode()
** This function changes the CLI statement attribute.
** In CLP, type
**            $chmod n
** where n is between 1 and 9
**
** $chmod 1 -- Dirty set by application
** $chmod 2 -- Dirty set by system (DB2 Everyplace default behavior)
** $chmod 3 -- Read include logical deleted records
** $chmod 4 -- Read only normal records
** $chmod 5 -- Read deleted record and set physical delete
** $chmod 6 -- Read normal record and set logical delete
** $chmod 7 -- Dirty set by application and read deleted records
** $chmod 8 -- Dirty set by system and read normal records
** $chmod 9 -- Dirty set by application and set physical delete (UDB behavior)
**
********************************************************************************/
SQLRETURN CLPSetStmtMode(CLPState *clpstate)
{
    SQLHSTMT hstmt = clpstate->hstmt;
    SQLRETURN rc = SQL_SUCCESS;

    //  check mode to set
    switch(clpstate->stmtCtrl)
    {
    case 1:
    {
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_DIRTYBIT_SET_MODE, (SQLPOINTER) SQL_DIRTYBIT_SET_BY_APPLICATION, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
                  
        //print1(clpstate,  (SQLTCHAR*)_TEXT("===>Dirty set by application\n"));
                                
        break;
    };
    
    case 2:
    {
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_DIRTYBIT_SET_MODE, (SQLPOINTER) SQL_DIRTYBIT_SET_BY_SYSTEM, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
  
        //print1(clpstate, (SQLTCHAR*)_TEXT("===>Dirty set by system\n"));

        break;
    };
    
    case 3:
    {
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_READ_MODE, (SQLPOINTER) SQL_READ_INCLUDE_MARKED_DELETE, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
    
        //print1(clpstate, (SQLTCHAR*)_TEXT("===>Read include logical deleted records\n"));
                
        break;
    };
    
    case 4:
    {
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_READ_MODE, (SQLPOINTER) SQL_READ_EXCLUDE_MARKED_DELETE, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
                 
        //print1(clpstate, (SQLTCHAR*)_TEXT("===>Read only normal records\n"));
                
        break;
    };
    
    case 5:
    { 
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_READ_MODE, (SQLPOINTER) SQL_READ_INCLUDE_MARKED_DELETE, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_DELETE_MODE, (SQLPOINTER) SQL_DELETE_PHYSICAL_REMOVE, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
    
        //print1(clpstate, (SQLTCHAR*)_TEXT("===>Read deleted record and set physical delete\n"));
                
        break;
    };
    
    case 6:
    {
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_READ_MODE, (SQLPOINTER) SQL_READ_EXCLUDE_MARKED_DELETE, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_DELETE_MODE, (SQLPOINTER) SQL_DELETE_MARK_ONLY, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
    
        //print1(clpstate, (SQLTCHAR*)_TEXT("===>Read normal record and set logical delete\n"));
                
        break;
    };
    
    case 7:
    {
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_DIRTYBIT_SET_MODE, (SQLPOINTER) SQL_DIRTYBIT_SET_BY_APPLICATION, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_READ_MODE, (SQLPOINTER) SQL_READ_INCLUDE_MARKED_DELETE, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
    
        //print1(clpstate, (SQLTCHAR*)_TEXT("===>Dirty set by application and read deleted records\n"));
                
        break;
    };
    
    case 8:
    {
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_DIRTYBIT_SET_MODE, (SQLPOINTER) SQL_DIRTYBIT_SET_BY_SYSTEM, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_READ_MODE, (SQLPOINTER) SQL_READ_EXCLUDE_MARKED_DELETE, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
    
        //print1(clpstate, (SQLTCHAR*)_TEXT("===>Dirty set by system and read normal records\n"));
                
        break;
    };

    case 9:
    {
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_DIRTYBIT_SET_MODE, (SQLPOINTER) SQL_DIRTYBIT_SET_BY_APPLICATION, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
        rc = SQLSetStmtAttr(hstmt, SQL_ATTR_DELETE_MODE, (SQLPOINTER) SQL_DELETE_PHYSICAL_REMOVE, 0);
        if ((rc != SQL_SUCCESS) && (rc != SQL_SUCCESS_WITH_INFO))
            goto exit;
    
        break;
    };

    default:
        break;
    };

    // mode set already at this point, set mode to UNCHANGED.
    clpstate->stmtCtrl = DB2eCLP_STMTATTR_UNCHANGED;

 exit:
    return rc;
}

// 03192001ctl begin
/********************************************************************************
 ** CLPStartEngineOnPath()
 **   This function starts the DB2 Everywhere engine and allocates
 **   the environment, connection, and statement handles.
********************************************************************************/
SQLRETURN CLPStartEngineOnPath(CLPState *clpstate, SQLTCHAR *server, SQLTCHAR *uid, SQLTCHAR *pwd)
{
    SQLHENV               henv;   
    SQLHDBC               hdbc;
    SQLHSTMT          hstmt;
    SQLRETURN     rc;
	int i = 0; //d4254qz
	
	//d4254qz 
	clpstate->mstmtno = 0;
	for(i = 0; i <20; i++)
    {
	   clpstate->mhstmt[i] = 0;
     clpstate->bindarray[i] = NULL;
    }

    clpstate->flag_hstmt = 0; //xh 092801
    clpstate->current_mhstmt = 0; //xh 092801

    // try to allocate the environment handle
    rc = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &henv);
    if (rc)
    { // if allocation fails
        return rc;
    }
    else
        clpstate->henv = henv;

    // try to allocate the database connection handle
    rc = SQLAllocHandle(SQL_HANDLE_DBC, henv, &hdbc);
    if (SQL_SUCCESS != rc)
    { // if allocation fails
        SQLFreeHandle(SQL_HANDLE_ENV, henv);
        return rc;
    }
    else
        clpstate->hdbc = hdbc;

    // start the connection to the database
    rc = SQLConnect(hdbc, server, SQL_NTS, uid, SQL_NTS, pwd, SQL_NTS);
    if (SQL_SUCCESS != rc)
    { // if connecting fails
        SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
        SQLFreeHandle(SQL_HANDLE_ENV, henv); 
        return rc;
    }

    // try to allocate the statement handle
    rc = SQLAllocHandle(SQL_HANDLE_STMT, hdbc, &hstmt);
    if (SQL_SUCCESS != rc)
    { // if allocation fails
        SQLDisconnect(hdbc);
        SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
        SQLFreeHandle(SQL_HANDLE_ENV, henv); 
        return rc;
    }       
    else
        clpstate->hstmt = hstmt;

    clpstate->stmtCtrl = DB2eCLP_STMTATTR_UNCHANGED;

    return rc;
}


/********************************************************************************
** CLPStartEngine()
**   This function starts the DB2 Everywhere engine using default directory.
********************************************************************************/
SQLRETURN CLPStartEngine(CLPState *clpstate)
{
#if defined(oEPOC)
    return  CLPStartEngineOnPath(clpstate, (SQLTCHAR*)_TEXT("C:\\System\\Data\\DB2e\\"), 
                                 (SQLTCHAR*)_TEXT("user"), (SQLTCHAR*)_TEXT("passwd"));
#else 
    return CLPStartEngineOnPath(clpstate, (SQLTCHAR*)_TEXT("server"), 
                                (SQLTCHAR*)_TEXT("user"), (SQLTCHAR*)_TEXT("passwd"));
#endif
}
// 03192001ctl end

/********************************************************************************
** CLPStopEngine
**   This function stops the DB2 Everywhere engine and cleans up
**   all environment, connection, and statement handles.
********************************************************************************/
void CLPStopEngine(CLPState *clpstate)
{
  int i,j;  
     for(j =0; j < 20; j++)
        {
        if (clpstate->bindarray[j] != NULL)
          {
          //paramvalue = clpstate->bindarray[j];
          if(clpstate->bindarray[j]){
           for(i=0; i<10; i++)
            {
            if (clpstate->bindarray[j][i]){
              free(clpstate->bindarray[j][i]);
              clpstate->bindarray[j][i] = NULL;
              }
            }
           free(clpstate->bindarray[j]);
           clpstate->bindarray[j] = NULL;
           }
          }
        }
    SQLFreeHandle(SQL_HANDLE_STMT, clpstate->hstmt);
    SQLDisconnect(clpstate->hdbc);
    SQLFreeHandle(SQL_HANDLE_DBC, clpstate->hdbc);
    SQLFreeHandle(SQL_HANDLE_ENV, clpstate->henv); 
}

// 03192001ctl begin
/********************************************************************************
 ** CLPReconnect
 **   This function connects the DB2 Everywhere engine to a particular database,
 **   i.e., directory.  For example,
 **      CLP:> connect to c:/temp/;
 **   Note that '/' is used instead of '\', and the last character of the path
 **   is also '/'.
********************************************************************************/
SQLRETURN CLPReconnect(CLPState *clpstate)
{
    SQLRETURN      rc = SQL_SUCCESS;
    SQLTCHAR       connectFormat[] = _TEXT("CONNECT TO %s");
    SQLTCHAR       server[300];
    SQLTCHAR       *pstr, *bstr; 
    unsigned int   numFields;
    int            i;
  
    // convert the first two words to uppercase
    for (i = 0; i < 11; i++) clpstate->cmd[i] = oToUpper(clpstate->cmd[i]);

    // Get the directory name
/*#if !defined(NOSSCANF) //08022001kmt
    numFields = _stscanf(clpstate->cmd, connectFormat, server);
    if (numFields != 1)
    {
        rc = -1;
        goto exit;
    }
#else*/
    if(0 == _tcsncmp(clpstate->cmd, _TEXT("CONNECT TO"), 10))     
    {
        pstr = clpstate->cmd + 10;
        bstr = clpstate->cmd + _tcslen(clpstate->cmd);
        //remove spaces coming before the argument of 3rd argument (server)
        while((*pstr == ' ') && pstr < bstr)
            pstr++;
        _tcscpy(server, pstr);
    }
    else
    {
        rc = -1;
        goto exit;
    }
//#endif

    // disconnect from the previous connection
    CLPStopEngine(clpstate);

    // connect to the new directory
    rc = CLPStartEngineOnPath(clpstate, (SQLTCHAR *)server, 
                              (SQLTCHAR *)_TEXT("user"), 
                              (SQLTCHAR *)_TEXT("passwd"));
                 
 exit:
    return rc;
  
}
// 03192001ctl end


/********************************************************************************
** CLPExecStatement()
**   This function takes in a statement handle and
**   the SQL string and executes it.
********************************************************************************/
SQLRETURN CLPExecStatement(CLPState *clpstate)
{
    SQLHSTMT       hstmt = clpstate->hstmt;
    SQLRETURN      rc = SQL_SUCCESS;
    SQLSMALLINT    numCols = 0;
    SQLUINTEGER    j;
    SQLTCHAR      *sqlstmt = &(clpstate->cmd[0]);
    unsigned int   rowCount = 0;
    SQLTCHAR      *stringbuf = 0;
    SQLTCHAR      *s, *sptr, *buf, *blob;
    SQLTCHAR      *p;
    SQLTCHAR      *colnames = 0;
    long           str1len, x;
    SQLINTEGER     i, len;
    SQLUINTEGER    k;
#define FORTY 40
    SQLTCHAR       dec[FORTY];
    SQLUINTEGER    length;
    DataTypes     *datatypes = 0;
    short          y;
    SQLTCHAR       err_string[BUFSIZE];
    unsigned int   numFields; //03232001xh
  
    //  Executes the statement
    rc = SQLExecDirect(hstmt, clpstate->cmd, SQL_NTS);
    if (rc)
    {
        ErrorReport(clpstate);
        goto exit;
    }

    rc = SQLNumResultCols(hstmt, &numCols);
    if (rc)
    {
        ErrorReport(clpstate);
        goto exit;
    }
            
    if (numCols)          // Assume that SELECT statement has been executed
    {
        // Allocate struct to store the data types of all columns
        datatypes = (DataTypes *) malloc(sizeof(DataTypes) * numCols);
        if (datatypes == 0)
        {
            print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_INTERNAL_ERROR2) );
            rc = -1;
            goto exit;
        }

        // Get the data type for each output column.
        // Compute the length required for the buffer and allocate memory.
        length = 0;
        for (i = 0; i < numCols; i++) 
        {
            SQLDescribeCol(hstmt, ( SQLSMALLINT ) (i + 1),
                           datatypes[i].colname,
                           sizeof(datatypes[i].colname),
                           (short *)&(datatypes[i].colnamelen),
                           (short *)&(datatypes[i].coltype),
                           &(datatypes[i].collen),
                           (short *)&(datatypes[i].scale),
                           (short *)&(datatypes[i].nullable));
            switch (datatypes[i].coltype)
            {
            case SQL_CHAR:
            case SQL_VARCHAR:
            case SQL_TYPE_DATE:
            case SQL_TYPE_TIME:
            case SQL_TYPE_TIMESTAMP:   //d2442akl

                length += (datatypes[i].colnamelen > datatypes[i].collen ?
                           datatypes[i].colnamelen : datatypes[i].collen);
                break;

            case SQL_DECIMAL:

                length += (datatypes[i].colnamelen > datatypes[i].collen+DECIMAL_POINT+SIGN ?
                           datatypes[i].colnamelen : datatypes[i].collen+DECIMAL_POINT+SIGN);
                break;

            case SQL_INTEGER:
            case SQL_SMALLINT:

                length += (datatypes[i].colnamelen > datatypes[i].collen+SIGN ?
                           datatypes[i].colnamelen : datatypes[i].collen+SIGN);
                break;

            case SQL_BLOB:

                if (datatypes[i].collen > 50)
                    length += 150 + 2;
                else length += (datatypes[i].colnamelen > datatypes[i].collen*3 ?
                                datatypes[i].colnamelen : datatypes[i].collen*3) + 2;
                break;

            default:

                length += (datatypes[i].colnamelen > datatypes[i].collen ?
                           datatypes[i].colnamelen : datatypes[i].collen);
                break;
            }

            length++; // for a ' '
        }

        // allocate memory accordingly
        length += 200; // for safety, add 200 more bytes --- old:20
        colnames  = (SQLTCHAR*)malloc(length * sizeof(SQLTCHAR));
        stringbuf = (SQLTCHAR*)malloc(length * sizeof(SQLTCHAR));
        if (colnames == 0 || stringbuf == 0)
        {
            print1 (clpstate, (SQLTCHAR*)_TEXT(CLP_INTERNAL_ERROR2) );
            rc = -1;
            goto exit;
        }

        // We echo the column names and dashes only when it is in INTERACT mode.
//d4182qz #if !defined(oWINCE) //&& !defined(oPALMOS) d2786akl commented out palmos part
        if (TSTBIT(clpstate->mode,CLP_INTERACT))
        {
            p = colnames;
//#ifndef oWINCE //09282001jgu	//d4486jgu comment out for regression test
            *p++ = '\n';
//#endif
            s = stringbuf;
            for (i = 0; i < numCols; i++) 
            {
                switch (datatypes[i].coltype)
                {
                case SQL_CHAR:
                case SQL_VARCHAR:
                case SQL_TYPE_DATE:
                case SQL_TYPE_TIME:
                case SQL_TYPE_TIMESTAMP:   //d2442akl
                                  
                    k = (datatypes[i].colnamelen > datatypes[i].collen ?
                         datatypes[i].colnamelen : datatypes[i].collen);
                    break;
                                  
                case SQL_DECIMAL:
                                  
                    k = (datatypes[i].colnamelen > datatypes[i].collen+DECIMAL_POINT+SIGN ?
                         datatypes[i].colnamelen : datatypes[i].collen+DECIMAL_POINT+SIGN);
                    break;
                                  
                case SQL_INTEGER:
                case SQL_SMALLINT:
                                  
                    k = (datatypes[i].colnamelen > datatypes[i].collen+SIGN ?
                         datatypes[i].colnamelen : datatypes[i].collen+SIGN);
                    break;
                                  
                case SQL_BLOB:
                                  
                    if (datatypes[i].collen > 50)
                        k = 150 + 2;
                    else k = (datatypes[i].colnamelen > datatypes[i].collen*3 ?
                              datatypes[i].colnamelen : datatypes[i].collen*3) + 2;
                    break;
                                  
                default:
                                  
                    k = (datatypes[i].colnamelen > datatypes[i].collen ?
                         datatypes[i].colnamelen : datatypes[i].collen);
                    break;
                }
                          
                for (j = 0; j < k; j++) *s++ = '-';
                *s++ = ' ';
                          
                _tcsncpy(p, datatypes[i].colname, datatypes[i].colnamelen);
                p += datatypes[i].colnamelen;
                for (j = datatypes[i].colnamelen; j < k; j++) *p++ = ' ';
                *p++ = ' ';
                          
            }

            s--;
            *s++ = 0;
            p--;
            *p++ = 0;
            writeOut(clpstate, colnames);
            writeOut(clpstate, stringbuf);
                
        }
//d4182qz #endif

        // Loop through the enable result table, one row at a time and output the data
        while (SQL_SUCCESS == (rc = SQLFetch(hstmt)))
        {
            rowCount++;
            s = stringbuf;
            
            for (i=0; i < numCols; i++) 
            {
                // add a comma as a separator for EXPORT
                if (TSTBIT(clpstate->mode,CLP_EXPORT) && i)
                {
                    *s++ = ',';
                }
                          
                switch (datatypes[i].coltype)
                {
                case SQL_CHAR:
                case SQL_VARCHAR:
                case SQL_TYPE_TIMESTAMP:
        
                    len = datatypes[i].collen + 2;
                    buf = sptr = (SQLTCHAR*)malloc(len * sizeof(SQLTCHAR)); // workspace to add escape char
                    memset(sptr, 0, len * sizeof(SQLTCHAR));
                    rc = SQLGetData(hstmt, i+1, SQL_C_CHAR, buf, len*sizeof(SQLTCHAR), &str1len);
                    if (rc) // 07232001ctl
                    {
                        ErrorReport(clpstate);
                        goto exit;
                    }
                    if (str1len > 0) // if str1len is less than 0, the result becomes 2147483647 after division in WinCE.
                        str1len /= sizeof(SQLTCHAR);
                    if (TSTBIT(clpstate->mode,CLP_EXPORT))
                    {// export mode
                        if (str1len >= 0)
                        { // maybe it is an empty string
                            *s++ = '"'; // Adds a leading double quote (assuming value not NULL)
                            while (str1len > 0)
                            {
                                *s++ = *sptr;
                                if (*sptr == '"') // add a '"' if we see a '"'
                                    *s++ = '"';
                                sptr++;
                                str1len--;
                            }
                            *s++ = '"'; // Adds a trailing double quote
                        }
                    }
                    else
                    {
                        k = (datatypes[i].colnamelen > datatypes[i].collen ?
                             datatypes[i].colnamelen : datatypes[i].collen);
                        if (str1len < 0) 
                        { // null
                            *s++ = '-'; 
                            for (j = 1; j < k; j++) *s++ = ' '; 
                        } 
                        else 
                        {
                            if (str1len > 0)
                            {
                                _tcsncpy(s, buf, str1len);
                                s += str1len;
                            }
                            for (j = str1len; j < k; j++) *s++ = ' '; 
                        }
                    }
                    free(buf);

                    break;

                case SQL_TYPE_DATE:
                case SQL_TYPE_TIME:

                    rc = SQLGetData(hstmt, i+1, SQL_C_CHAR, s, (length+s-stringbuf)*sizeof(SQLTCHAR), &str1len);
                    if (rc) // 07232001ctl
                    {
                        ErrorReport(clpstate);
                        goto exit;
                    }
                    if (str1len > 0) // if str1len is less than 0, the result becomes 2147483647 after division in WinCE.
                        str1len /= sizeof(SQLTCHAR);
                    if (TSTBIT(clpstate->mode,CLP_EXPORT))
                    {// export mode
                        if (str1len > 0)
                        {
                            if (datatypes[i].coltype == SQL_TYPE_DATE)
                                s += DB2eDateToCSVDate(s);
                            else
                                s += DB2eTimeToCSVTime(s);
                        }
                    }
                    else
                    {
                        k = (datatypes[i].colnamelen > datatypes[i].collen ?
                             datatypes[i].colnamelen : datatypes[i].collen);
                        if (str1len < 0) 
                        { // null
                            *s++ = '-'; 
                            for (j = 1; j < k; j++) *s++ = ' '; 
                        } 
                        else 
                        {
                            s += str1len;
                            for (j = str1len; j < k; j++) *s++ = ' '; 
                        }
                    }

                    break;

                case SQL_DECIMAL:

                    rc = SQLGetData(hstmt, i+1, SQL_C_CHAR, dec, FORTY*sizeof(SQLTCHAR), &str1len); // 07132001ctl
                    if (rc) // 07232001ctl
                    {
                        ErrorReport(clpstate);
                        goto exit;
                    }
                    if (str1len > 0) // if str1len is less than 0, the result becomes 2147483647 after division in WinCE.
                        str1len /= sizeof(SQLTCHAR);
                    if (TSTBIT(clpstate->mode,CLP_EXPORT))
                    {// export mode
                        if (str1len > 0)
                        {
                            _tcsncpy(s, dec, str1len);
                            s += str1len;
                        }
                    }
                    else
                    {
                        k = (datatypes[i].colnamelen > datatypes[i].collen+DECIMAL_POINT+SIGN ?
                             datatypes[i].colnamelen : datatypes[i].collen+DECIMAL_POINT+SIGN);
                        if (str1len < 0) 
                        { // null
                            for (j = 1; j < k; j++) *s++ = ' '; 
                            *s++ = '-'; 
                        } 
                        else 
                        {
                            for (j = str1len; j < k; j++) *s++ = ' ';
                            _tcsncpy(s, dec, str1len);
                            s += str1len;
                        }
                    }

                    break;

                case SQL_SMALLINT:

                    rc = SQLGetData(hstmt, i+1, SQL_C_SHORT, &y, sizeof(y), (long *) &str1len);
                    if (rc) // 07232001ctl
                    {
                        ErrorReport(clpstate);
                        goto exit;
                    }
                    if (TSTBIT(clpstate->mode,CLP_EXPORT))
                    {// export mode
                        if (str1len >= 0)
                        {
                            //Palm CLP begin
#if defined(oPALMOS)
                            StrIToA((char*)s,y);
#else 
                            _stprintf(s,_TEXT("%d"),y);
#endif
                            //Palm CLP end 
                            s += _tcslen(s);
                        }
                    }
                    else
                    {
                        k = (datatypes[i].colnamelen > datatypes[i].collen+SIGN ?
                             datatypes[i].colnamelen : datatypes[i].collen+SIGN);
                        if (str1len < 0) 
                        {
                            for (j = 1; j < k; j++) *s++ = ' '; 
                            *s++ = '-'; 
                        } 
                        else 
                        {
                            for (j = 6; j < k; j++) *s++ = ' '; 
#if defined(oEPOC) && defined(UNICODE)
                            _stprintf((char *)s, _TEXT("%6d"), (long)y); //08022001kmt
#else
                            _stprintf((char *)s, _TEXT("%6ld"), (long)y); //07132001ctl
#endif
                            s += _tcslen(s);
                        }
                    }

                    break;

                case SQL_INTEGER:

                    rc = SQLGetData(hstmt, i+1, SQL_C_LONG, &x, sizeof(x), (long *) &str1len);
                    if (rc) // 07232001ctl
                    {
                        ErrorReport(clpstate);
                        goto exit;
                    }
                    if (TSTBIT(clpstate->mode,CLP_EXPORT))
                    {// export mode
 
                        if (str1len >= 0)
                        {
                            //Palm CLP begin
#if defined(oPALMOS)
                            StrIToA((char*)s, x);
#else 
                            _stprintf(s,_TEXT("%d"),x);
#endif
                            //Palm CLP end 
                            s += _tcslen(s);
                        }
                    }
                    else
                    {
                        k = (datatypes[i].colnamelen > datatypes[i].collen+SIGN ?
                             datatypes[i].colnamelen : datatypes[i].collen+SIGN);
                        if (str1len < 0) 
                        {
                            for (j = 1; j < k; j++) *s++ = ' '; 
                            *s++ = '-'; 
                        } 
                        else 
                        {
                            for (j = 11; j < k; j++) *s++ = ' '; 
#if defined(oEPOC) && defined(UNICODE)
                            _stprintf((char *)s,_TEXT("%11d"),(long)x); //08022001kmt
#else
                            _stprintf((char *)s,_TEXT("%11ld"),(long)x); // 07132001ctl
#endif
                            s += _tcslen(s);
                        }
                    }

                    break;

                case SQL_BLOB:
                    buf = sptr = (SQLTCHAR *)malloc(50); // print only 50 bytes max
                    memset(sptr, 0, 50);
                    rc = SQLGetData(hstmt, i+1, SQL_C_BINARY, buf, 50, &str1len);
                    if (rc) // 07232001ctl
                    {
                        ErrorReport(clpstate);
                        goto exit;
                    }
                                 
                    if (TSTBIT(clpstate->mode,CLP_EXPORT))
                    { // export mode
                        // do nothing
                    }
                    else
                    {
                        if (datatypes[i].collen > 50)
                            k = 150 + 2;
                        else k = (datatypes[i].colnamelen > datatypes[i].collen*3 ?
                                  datatypes[i].colnamelen : datatypes[i].collen*3) + 2;
                        if (str1len < 0) 
                        {
                            for (j = 1; j < k; j++) *s++ = ' '; 
                            *s++ = '-';
                        } 
                        else 
                        {
                            //*s++ = '*';
                            int bloblen = (50>datatypes[i].collen)?datatypes[i].collen:50;
                            SQLCHAR *ch = (SQLCHAR *)buf;
                            SQLCHAR tempbuf[16];
                                                  
                            int t = 0;
                            *s++ = '0';
                            *s++ = 'x';
                            for (t=0; t<bloblen; t++){                             
#if defined(oPALMOS)
                                _stprintf((char *)tempbuf, _TEXT("%02x"), *(ch+t));
                                StrNCopy((char*)s, (char*)tempbuf+6, 2); // copy only two bytes
                                                        
#else 
                                _stprintf((char *)s, _TEXT("%02x"), *(ch+t));
#endif
                                s+=2;
                                *s++=' ';
                            } 

                            for (j = bloblen*3 + 2; j < k; j++) *s++ = ' '; 
                                                  
                        }
                    }
                    free(buf);

                    break;

                default:

                    break;
                }

                if (TSTBIT(clpstate->mode,CLP_INTERACT))
                    *s++ = ' ';
            } // for

            if (TSTBIT(clpstate->mode,CLP_INTERACT))
                s--;
            *s = 0;
            writeOut(clpstate, stringbuf);
        } // End of WHILE loop to fetch all records

        if (TSTBIT(clpstate->mode,CLP_INTERACT))
        {
            if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO || rc == SQL_NO_DATA_FOUND)
            {
                _stprintf((char*)err_string, (char*)_TEXT(CLP_MSG_ROW_COUNT), rowCount);
                writeOut(clpstate, err_string);
                rc = SQL_SUCCESS; // 04022001ctl
            }
        }
        else if (TSTBIT(clpstate->mode,CLP_EXPORT))
        {
            _stprintf((char*)err_string, (char*)_TEXT(CLP_EXPORT_OK), rowCount);
            print1(clpstate, err_string);
        }

        if (rc != SQL_SUCCESS && rc != SQL_NO_DATA_FOUND)
        {
            ErrorReport(clpstate);
            goto exit;
        }
    }
    else
    {
        if (TSTBIT(clpstate->mode,CLP_INTERACT))
            writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
    }


 exit:
    if (datatypes) free(datatypes);
    if (colnames)  free(colnames);
    if (stringbuf) free(stringbuf);

    return rc;
}



/********************************************************************************
** DB2eImport()
********************************************************************************/
SQLRETURN DB2eImport(CLPState *clpstate)
{
    SQLRETURN      rc = SQL_SUCCESS;
    SQLHSTMT       hstmt = clpstate->hstmt;
    SQLTCHAR       sqlstmt[BUFSIZE];        // Table name is 18 characters max, so BUFSIZE is enough
    SQLTCHAR       insert_stmt[BUFSIZE];    // Table name is 18 characters max, so BUFSIZE is enough
    SQLTCHAR      *table_name = 0;
    SQLTCHAR       importFormat[] = _TEXT("IMPORT FROM %s");
    SQLTCHAR      *importSQL, *cptr, *bkptr, *sptr;
    SQLTCHAR       inPath[ID_LEN];
    SQLTCHAR       inFile[ID_LEN];
    SQLTCHAR       astr[ID_LEN];//Palm CLP, use as internal buffer
    SQLTCHAR      *ofPart;//Palm CLP
    SQLTCHAR      *pTemp;
    SQLTCHAR      *wkBuf    = 0; // workspace for massaging input data
    SQLTCHAR      *pCSVData = 0;
    SQLTCHAR      *readBuf  = 0, *pBuf;
    SQLINTEGER    *sql_null_data = 0; // an array of nullability indicator or the length of data in input buffer
    SQLTCHAR      *date;
    SQLTCHAR      *time;
    SQLTCHAR      *timestamp; // d03212001ctl
    SQLTCHAR       c;
    SQLSMALLINT    numcols = 0;
    unsigned int   numFields = 0;
    int            i, j, int_value, k=-1, l;//k,l is added for Palm CLP
    int            length;
    DataTypes     *datatypes = 0;
    long          *int_array = 0;  // d11182000ctl
    int            nbRec = 0;
    int            buflen = 0;
    unsigned int   x[7]; // d03212001ctl
#define EIGHTY 80
    SQLTCHAR       buffer[EIGHTY];
    char           tempbuffer[EIGHTY];
    int            nb_read;
    SQLUINTEGER    idx = sizeof(SQLTCHAR);
    SQLTCHAR       err_string[BUFSIZE];
    int n_tag; 
#define Isdigit(a) ((('0'<= a )&&(a <= '9')) ? 1 : 0)//Palm CLP
    int d_tag;     //Palm CLP
    int            string_read = 0;          // 04022001ctl
    int            double_quoted_string = 0; // 04022001ctl
    int chsz; //d3041  
    char ch; //d3041

    // convert the first two words to uppercase
    for (i = 0; i < 11; i++) clpstate->cmd[i] = oToUpper(clpstate->cmd[i]);

    // Get the import filename
//#if defined(NOSSCANF) //Palm CLP begin //08022001kmt
    //Parsing the input to get correct file name
    _tcscpy((SQLTCHAR*)inPath, (SQLTCHAR*)(clpstate->cmd + 12));
    j = _tcslen((SQLTCHAR*)inPath);
    for(i = 0; i < j; i ++) inPath[i] = oToUpper(inPath[i]); 
    ofPart = (SQLTCHAR*)_tcsstr((char*)inPath, (char*)_TEXT("OF"));
    length = _tcslen((SQLTCHAR*)inPath) - _tcslen((SQLTCHAR*)ofPart) - 1;// -1 if for the ending space
    memset(inFile, 0, ID_LEN);
    _tcsncpy((SQLTCHAR*)inFile, (SQLTCHAR*)(clpstate->cmd + 12), length);
    inFile[length] = 0; //null terminate it
    numFields = 1;
//#else//Palm CLP end
  //  numFields = _stscanf(clpstate->cmd, importFormat, inFile);
//#endif
  
    if (numFields == 1)
    {
#if defined oEPOC
        _tcscpy(inPath, _TEXT("C:\\System\\Data\\DB2e\\"));
        _tcscat(inPath, inFile);
#elif defined(WIN32U) // 07132001ctl
        memset(inPath, 0, ID_LEN);
        WideCharToMultiByte(CP_ACP,
                            0, 
                            (LPTSTR) inFile,
                            -1,
                            inPath, 
                            ID_LEN,
                            NULL,
                            NULL);
#else
        _tcscpy(inPath, inFile);
#endif
        if ( inPath != 0 )
        { 
#if defined(oPALMOS)//Palm CLP begin
            //o_Fopen will only return NULL when there is failure
            clpstate->fpIN = o_Fopen((char*)inPath, (char*)_TEXT("rb"));   
            if (clpstate->fpIN == NULL)
            {
                print2 ( clpstate, (SQLTCHAR*)_TEXT(CLP_IMPORT_FILE), inPath);
                rc = -1;
                goto exit;
            }//Palm CLP end
#else
            clpstate->fpIN = o_Fopen(inPath, _TEXT("rb"));   
            if (clpstate->fpIN == 0 ||clpstate->fpIN == -1)
            {
                print2 ( clpstate, (SQLTCHAR*)_TEXT(CLP_IMPORT_FILE), inPath);
                rc = -1;
                goto exit;
            }
#endif

            // Get the import table name
            importSQL = clpstate->cmd + _tcslen(_TEXT("IMPORT FROM ")) + _tcslen(inFile) + 1;
                  
            // convert the following two words to uppercase and compare
            // we support only DEL format
            j = _tcslen(_TEXT("OF DEL INSERT INTO "));
            for (i = 0; i < j; i++) importSQL[i] = oToUpper(importSQL[i]);
            if (0 == _tcsncmp(importSQL, _TEXT("OF DEL INSERT INTO "), j))
            {
                // get the table name and convert it to uppercase
                table_name = importSQL + j;
//                        for (i = 0; i < _tcslen(table_name); i++) table_name[i] = oToUpper(table_name[i]);
#if defined(oWINCE) //d3041
                for (i = 0; i < _tcslen(table_name); i++) { //d3041
                    chsz=nTxCharSz((Uchar)table_name[i]);
                    if (chsz==1)
                        table_name[i] = oToUpper(table_name[i]);
                    else
                        i+=chsz-1;     
                }
#else
                for (i = 0; i < _tcslen(table_name); i++) { //d3041
                    ch=(char)table_name[i];
                    if (ch>0)
                        table_name[i] = oToUpper(table_name[i]);
                    else
                        i++;   
                }
#endif                  
                memset(sqlstmt, 0, BUFSIZE);
                // Construct a SQL to find out the number of columns in the import table
                _tcscpy(sqlstmt, SQL_SELECT);
                _tcscat((char*)sqlstmt, (char*)table_name);
                          
                // Executes the statement
                rc = SQLExecDirect(hstmt, sqlstmt, SQL_NTS); 
                if (rc)
                {
                    ErrorReport(clpstate);
                    goto exit;
                }

                // get the number of columns
                rc = SQLNumResultCols(hstmt, &numcols);
                if (rc)
                {
                    ErrorReport(clpstate);
                    goto exit;
                }
            
                if (numcols)
                {
                    // Allocate struct to store the data type of each column and nullability columns
                    datatypes = (DataTypes *)malloc(sizeof(DataTypes) * numcols);
                    sql_null_data = (long *)malloc(sizeof(SQLINTEGER) * numcols);
                    if (datatypes == 0 || sql_null_data == 0)
                    {
                        print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_INTERNAL_ERROR2) );
                        rc = -1;
                        goto exit;
                    }

                    // Keep the data type
                    length = 0; // compute the total length; use for allocating buffer
                    for (i = 0; i < numcols; i++) 
                    {
                        SQLDescribeCol(hstmt, ( SQLSMALLINT ) (i + 1),
                                       datatypes[i].colname,
                                       sizeof(datatypes[i].colname),
                                       (short *)&(datatypes[i].colnamelen),
                                       (short *)&(datatypes[i].coltype),
                                       &(datatypes[i].collen),
                                       (short *)&(datatypes[i].scale),
                                       (short *)&(datatypes[i].nullable));

                        // add 10 spaces for each columns: '+', '-', '"', ','
                        if (datatypes[i].coltype == SQL_CHAR ||
                            datatypes[i].coltype == SQL_VARCHAR)
                        {// need to multiply by 2 because of double quote
                            length += (2*datatypes[i].collen*sizeof(SQLTCHAR) + TEN_SPACES);
                        }
                        else if (datatypes[i].coltype == SQL_INTEGER ||
                                 datatypes[i].coltype == SQL_SMALLINT)
                        {
                            length += (sizeof(long) + TEN_SPACES);
                        }
                        else if (datatypes[i].coltype == SQL_DECIMAL)
                        {
                            length += (datatypes[i].collen+DECIMAL_POINT+SIGN + TEN_SPACES);
                        }
                        else
                        {
                            length += (datatypes[i].collen + TEN_SPACES);
                        }

                        if (datatypes[i].coltype == SQL_BLOB)
                        {
                            print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_IMPORT_BLOB) );
                            goto exit;
                        }
                    }

                    // Allocate memory for reading a line from the input file
                    // and memory for bind parameter
                    readBuf  = (SQLTCHAR*)malloc(length * sizeof(SQLTCHAR));
                    wkBuf    = (SQLTCHAR*)malloc(length * sizeof(SQLTCHAR));
                    pCSVData = (SQLTCHAR*)malloc(length * sizeof(SQLTCHAR));
                    int_array = (long *)malloc(numcols * sizeof(long));  // d11182000ctl
                    if (readBuf == 0 || pCSVData == 0 || wkBuf == 0)
                    {
                        print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_INTERNAL_ERROR2) );
                        rc = -1;
                        goto exit;
                    }

                    // Construct the INSERT statement
                    //   INSERT INTO table-name VALUES (?, ....,?)
                    cptr = insert_stmt;
                    memset(insert_stmt, 0, BUFSIZE);
                    _tcscpy(cptr, SQL_INSERT); cptr += _tcslen(SQL_INSERT);
                    _tcscpy(cptr, table_name); cptr += _tcslen(table_name);
                    _tcscpy(cptr, SQL_VALUES); cptr += _tcslen(SQL_VALUES);
                    for (i = 0; i < numcols; i++)
                    {
                        *cptr++ = '?';
                        *cptr++ = ',';
                    }
                    cptr--; // backspace once
                    *cptr++ =  ')';
                    *cptr = '\0'; // NTS

                    // Prepare an INSERT statement
                    rc = SQLPrepare(hstmt, insert_stmt, SQL_NTS); 
                    if (rc)
                    {
                        ErrorReport(clpstate);
                        goto exit;
                    }

                    // Bind the column buffers
                    pTemp = pCSVData;
                    for (i = 0; i < numcols; i++) 
                    {
                        switch (datatypes[i].coltype)
                        {
                        case SQL_INTEGER:
                            // d11182000ctl begin
                            // lptr = (long*) pTemp;
                            // pTemp += (sizeof(long) + EXTRA_SPACE); // add one more space
                            // d11182000ctl end
                            rc = SQLBindParameter(hstmt,
                                                  i+1, 
                                                  SQL_PARAM_INPUT, 
                                                  SQL_C_LONG,                // C data type of this parameter
                                                  SQL_INTEGER,               // SQL data type of this parameter
                                                  0,                         // Precision. Not used.
                                                  0,                         // Scale. Not used.
                                                  &(int_array[i]),       // Parameter value // d11182000ctl
                                                  0,                     // Length of the parameter value (not used)
                                                  &(sql_null_data[i]));  // nullability
                            if (rc)
                            {
                                ErrorReport(clpstate);
                                goto exit;
                            }
                                                  
                            break;

                        case SQL_SMALLINT:
                            // d11182000ctl begin
                            // lptr = (long*) pTemp;
                            // pTemp += (sizeof(long) + EXTRA_SPACE); // add one more space
                            // d11182000ctl end
                            rc = SQLBindParameter(hstmt,
                                                  i+1, 
                                                  SQL_PARAM_INPUT, 
                                                  SQL_C_LONG,                // C data type of this parameter
                                                  SQL_SMALLINT,              // SQL data type of this parameter
                                                  0,                         // Precision. Not used.
                                                  0,                         // Scale. Not used.
                                                  &(int_array[i]),       // Parameter value // d11182000ctl
                                                  0,                     // Length of the parameter value (not used)
                                                  &(sql_null_data[i]));  // nullability
                            if (rc)
                            {
                                ErrorReport(clpstate);
                                goto exit;
                            }

                            break;

                        case SQL_CHAR:

                            cptr = pTemp;
                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR); // add one more space
                            rc = SQLBindParameter(hstmt,
                                                  i+1, 
                                                  SQL_PARAM_INPUT, 
                                                  SQL_C_CHAR,                // C data type of this parameter
                                                  SQL_CHAR,                  // SQL data type of this parameter
                                                  0,                         // Precision. Not used.
                                                  0,                         // Scale. Not used.
                                                  cptr,                      // Parameter value 
                                                  0,                     // Length of the parameter value (not used)
                                                  &(sql_null_data[i]));  // nullability
                            if (rc)
                            {
                                ErrorReport(clpstate);
                                goto exit;
                            }

                            break;

                        case SQL_VARCHAR:

                            cptr = pTemp;
                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR); // add one more space
                            rc = SQLBindParameter(hstmt,
                                                  i+1, 
                                                  SQL_PARAM_INPUT, 
                                                  SQL_C_CHAR,                // C data type of this parameter
                                                  SQL_VARCHAR,               // SQL data type of this parameter
                                                  0,                         // Precision. Not used.
                                                  0,                         // Scale. Not used.
                                                  cptr,                      // Parameter value 
                                                  0,                     // Length of the parameter value (not used)
                                                  &(sql_null_data[i]));  // nullability
                            if (rc)
                            {
                                ErrorReport(clpstate);
                                goto exit;
                            }

                            break;

                        case SQL_DECIMAL:

                            cptr = pTemp;
                            pTemp += (datatypes[i].collen+DECIMAL_POINT+SIGN+EXTRA_SPACE)*sizeof(SQLTCHAR);
                            rc = SQLBindParameter(hstmt,
                                                  i+1, 
                                                  SQL_PARAM_INPUT, 
                                                  SQL_C_CHAR,                      // C data type of this parameter
                                                  SQL_DECIMAL,                     // SQL data type of this parameter
                                                  0,                               // Precision. Not used.
                                                  0,                               // Scale. Not used.
                                                  cptr,                            // Parameter value
                                                  0,                     // Length of the parameter value (not used)
                                                  &(sql_null_data[i]));    // nullability
                            if (rc)
                            {
                                ErrorReport(clpstate);
                                goto exit;
                            }

                            break;

                        case SQL_TYPE_DATE:

                            date = pTemp;
                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR); // MM-DD-YYYY
                            rc = SQLBindParameter(hstmt,
                                                  i+1, 
                                                  SQL_PARAM_INPUT, 
                                                  SQL_C_CHAR,                // C data type of this parameter
                                                  SQL_TYPE_DATE,             // SQL data type of this parameter
                                                  0,                         // Precision. Not used.
                                                  0,                         // Scale. Not used.
                                                  date,                      // Parameter value
                                                  0,                     // Length of the parameter value (not used)
                                                  &(sql_null_data[i]));  // nullability
                            if (rc)
                            {
                                ErrorReport(clpstate);
                                goto exit;
                            }

                            break;

                        case SQL_TYPE_TIME:

                            time = pTemp;
                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR); // HH:MM:SS
                            rc = SQLBindParameter(hstmt,
                                                  i+1, 
                                                  SQL_PARAM_INPUT, 
                                                  SQL_C_CHAR,                // C data type of this parameter
                                                  SQL_TYPE_TIME,             // SQL data type of this parameter
                                                  0,                         // Precision. Not used.
                                                  0,                         // Scale. Not used.
                                                  time,                      // Parameter value
                                                  0,                     // Length of the parameter value (not used)
                                                  &(sql_null_data[i]));  // nullability
                            if (rc)
                            {
                                ErrorReport(clpstate);
                                goto exit;
                            }

                            break;

                        case SQL_TYPE_TIMESTAMP: // d03212001ctl begin

                            timestamp = pTemp;
                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR); // MM-DD-YYYY
                            rc = SQLBindParameter(hstmt,
                                                  i+1, 
                                                  SQL_PARAM_INPUT, 
                                                  SQL_C_CHAR,                // C data type of this parameter
                                                  SQL_TYPE_TIMESTAMP,    // SQL data type of this parameter
                                                  0,                         // Precision. Not used.
                                                  0,                         // Scale. Not used.
                                                  timestamp,             // Parameter value
                                                  0,                     // Length of the parameter value (not used)
                                                  &(sql_null_data[i]));  // nullability
                            if (rc)
                            {
                                ErrorReport(clpstate);
                                goto exit;
                            }

                            break; // d03212001ctl end

                        default:
                            rc = -1; // unsupported data type
                            goto exit;
                        }
                    }

                    // Scan the input file.
                    // For each line, do:
                    //  1. convert to the proper format
                    //  2. store the data value in appropriate buffer
                    //  3. execute the insert statement
                    pBuf   = readBuf;
                    buflen = 0;
                    memset(pBuf, 0 , length * sizeof(SQLTCHAR));
                    nb_read = 1;
                    while ( /* !oFeof(clpstate->fpIN) d11182000ctl && */ nb_read > 0)
                    {
                        // read some bytes from the input file
                        // process this line; we may hit end-of-line                                    
                        memset(buffer,0,EIGHTY * sizeof(SQLTCHAR));
#if defined(WIN32U) // 07132001ctl convert to UCS-2
                        memset(tempbuffer, 0, EIGHTY);
                        nb_read = o_Fread(tempbuffer, sizeof(char), EIGHTY, clpstate->fpIN);
                        if (nb_read > 0)
                            MultiByteToWideChar(CP_ACP, 0, tempbuffer, -1, buffer, nb_read);
                        nb_read = nb_read << 1; // times 2
#else
                        //nb_read is the number of bytes read
                        nb_read = o_Fread(buffer, sizeof(SQLTCHAR), EIGHTY, clpstate->fpIN) * sizeof(SQLTCHAR);
#endif

                        // d11182000ctl begin
                        if (nb_read <= 0 && pBuf == readBuf)                                          
                            break;  // no more from the input file and no more line to process
                        else  if (nb_read <= 0 && pBuf != readBuf)
                        { // process the last line
                            nb_read = sizeof(SQLTCHAR);
                            buffer[0] = '\n';
                        }
                        // d11182000ctl end

                        n_tag = 0;
                        for (idx = 0; idx < nb_read/sizeof(SQLTCHAR); idx++) //loop here after SQLexecute
                        {
                            c = buffer[idx];
                            // if (EOF == c)
                            // break;

                            // ignore '\r'
                            if (c == '\r')
                                continue;
                                                
                            if((c == '\n') && (buflen == 0)
                               ||(c == '\n') && (n_tag == 1))
                                continue;

                            // end of line
                            if (c != '\n')
                            {
                                n_tag = 0;
                                // store the character
                                *pBuf++ = c;
                                if (++buflen >= length)
                                {
                                    print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_INTERNAL_ERROR1) );
                                    goto exit;
                                }
                            }
                            else

                            {                                                       
                                n_tag = 1;
                                // end of line
                                // goto the first character of the read buffer
                                pBuf = readBuf;
                                                          
                                // Copy and massage the input data to the designated column buffers
                                memset(pCSVData, 0 , length * sizeof(SQLTCHAR)); // clean the buffer
                                pTemp = pCSVData;
                                                          
                                // massage the input data and copy to the column buffers
                                memset(wkBuf, 0 , length * sizeof(SQLTCHAR)); // clean the workspace
                                cptr = wkBuf;
                                                          
                                // set the null indicators: by default, it is a null value
                                for (i = 0; i < numcols; i++)
                                {
                                    sql_null_data[i] = SQL_NULL_DATA; // null
                                }

                                i = 0;
                                string_read = 0;          // 04022001ctl
                                double_quoted_string = 0; // 04022001ctl
                                j = 0;
                                while (j < buflen) //end of an input line
                                { // 04022001ctl begin
                                    // a comma may be embedded inside a double-quoted string "ab,cd"
                                    // remove leading space
                                    if (*pBuf == ' ') {
                                        pBuf++;
                                        j++;
                                        continue;
                                    }

                                    // special handling of double quoted strings
                                    if ((datatypes[i].coltype == SQL_CHAR || datatypes[i].coltype == SQL_VARCHAR) &&
                                        *pBuf == '"' && // starting with a double quote
                                        double_quoted_string == 0) {
                                        // we have to see even number of '"'
                                        double_quoted_string = 1; // the string starts with '"'
                                        //*cptr++ = *pBuf++;        // copy '"' to buffer //jca
                                        //j++; //jca

                                        while (j < buflen && double_quoted_string == 1) {

                                            *cptr++ = *pBuf++; // copy to buffer
                                            j++;

                                            if (*pBuf == '"' && j < buflen && *(pBuf+1) == '"') {
                                                *cptr++ = *pBuf++; // copy to buffer
                                                *cptr++ = *pBuf++; // copy to buffer
                                                j += 2;
                                            }
                                            else if (*pBuf == '"') {
                                                *cptr++ = *pBuf++; // copy to buffer
                                                j++;
                                                double_quoted_string = 0;
                                            }
                                        }
                                    }

                                    // copy the data value
                                    while (j < buflen && *pBuf != ',') {
                                        *cptr++ = *pBuf++; // copy to buffer
                                        j++;
                                    }
                                    // 04022001ctl end

                                    if (j == buflen  /* end of line */ ||
                                        *pBuf == ',' /* separator */)
                                    { // found a field/column
                                        switch (datatypes[i].coltype)
                                        {
                                        case SQL_INTEGER:
                                        case SQL_SMALLINT:
                                                                                  
                                            if (cptr != wkBuf)
                                            { 
#if !defined(NOSSCANF) //08022001kmt
                                                if (1 == _stscanf(wkBuf, DB2e_intFormat, &int_value))
                                                {
                                                    // memcpy(pTemp, &int_value, sizeof(long));
                                                    int_array[i] = int_value; // d11182000ctl
                                                    sql_null_data[i] = sizeof(long); // length of input
                                                }
                                                else
                                                {
                                                    print1 ( clpstate, _TEXT(CLP_IMPORT_DATA) );
                                                    goto exit;
                                                }
                                                //Palm CLP begin              
#else
                                                // remove all leading blanks (not tabs!!! according to DB2/UDB)
                                                bkptr = wkBuf;
                                                while (bkptr < cptr && *bkptr == ' ')
                                                    bkptr++;
                                                                                  
                                                // remove all trailing blanks (not tabs!!! according to DB2/UDB)
                                                while (bkptr < cptr && *(cptr-1) == ' ')
                                                {
                                                    *(cptr-1) = '\0';
                                                    cptr--;
                                                }
                                                                                        
                                                //parsing input to get correct value  
                                                memset(astr, 0, ID_LEN);      
                                                if (cptr-bkptr)
                                                    _tcsncpy((SQLTCHAR*)astr, (SQLTCHAR*)bkptr, (cptr-bkptr)*sizeof(SQLTCHAR)); //08022001kmt
                                                astr[cptr-bkptr] = 0; //null terminate it     
                                                k = -1;
                                                k = _tcslen((SQLTCHAR*)astr); //08022001kmt
                                                d_tag = 1;
                                                if((!Isdigit(astr[0]))&&(astr[0]!='-')&&(astr[0]!='+')) //sign support
                                                    d_tag = 0;                
                                                                                                                                                  
                                                for(l=1; l<k; l++) //jca
                                                    if(!Isdigit(astr[l]))
                                                    {
                                                        d_tag = 0;
                                                        break;
                                                    }                                                                                                                                                                                        
                                                if ((k != -1)&&(d_tag))
                                                {
                                                    // memcpy(pTemp, &int_value, sizeof(long));
                                                    int_value = _tcsAtoI((char*)astr); //08022001kmt
                                                    int_array[i] = int_value; // d11182000ctl
                                                    sql_null_data[i] = sizeof(long); // length of input
                                                }
                                                else
                                                {
                                                    print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_IMPORT_DATA) );
                                                    goto exit;
                                                }               
#endif
                                                //Palm CLP end                
                                            }
                                                                                  
                                            // pTemp += (sizeof(long) + EXTRA_SPACE);
                                                                                  
                                            break;
                                                                                  
                                        case SQL_CHAR:
                                        case SQL_VARCHAR:

                                            // remove all leading blanks (not tabs!!! according to DB2/UDB)
                                            bkptr = wkBuf;
                                            while (bkptr < cptr && *bkptr == ' ')
                                                bkptr++;
                                                                                  
                                            // remove all trailing blanks (not tabs!!! according to DB2/UDB)
                                            while (bkptr < cptr && *(cptr-1) == ' ')
                                            {
                                                *(cptr-1) = '\0';
                                                cptr--;
                                            }
                                                                                  
                                            if (bkptr < cptr)
                                            { // we have gotten a non-null string
                                                if (*bkptr != '"') 
                                                { // first char is not double-quote
                                                    // i.e., not a double-quoted string
                                                    if (cptr-bkptr)
                                                        memcpy(pTemp, bkptr, (cptr-bkptr)*sizeof(SQLTCHAR)); // d11182000ctl
                                                    sql_null_data[i] = _tcslen(pTemp)*sizeof(SQLTCHAR);
                                                }
                                                else
                                                { // a double-quoted string
                                                    bkptr++; // pass the leading double quote
                                                    sptr = pTemp;
                                                    while (bkptr < cptr)
                                                    {
                                                        if (*bkptr == '"')
                                                        {
                                                            if (bkptr+1 < cptr && *(bkptr+1) == '"')
                                                            {
                                                                bkptr++; // skip the escape char '"'
                                                                *sptr++ = *bkptr++;
                                                            }
                                                            else break;
                                                        }
                                                        else *sptr++ = *bkptr++;
                                                    }
                                                    sql_null_data[i] = _tcslen(pTemp)*sizeof(SQLTCHAR);
                                                }
                                            }

                                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR);
                                                                          
                                            break;
                                                                          
                                        case SQL_DECIMAL:
                                                                                  
                                            // remove all leading blanks or tabs
                                            bkptr = wkBuf;
                                            while (bkptr < cptr && *bkptr <= ' ')
                                                bkptr++;
                                            if (cptr != wkBuf && bkptr < cptr)
                                            {
                                                // remove all trailing blanks or tabs
                                                while (bkptr < cptr && *(cptr-1) <= ' ')
                                                {
                                                    *(cptr-1) = '\0';
                                                    cptr--;
                                                }
                                                memcpy(pTemp, bkptr,
                                                       (datatypes[i].collen+DECIMAL_POINT+SIGN)*sizeof(SQLTCHAR));
                                                sql_null_data[i] = _tcslen(pTemp)*sizeof(SQLTCHAR);
                                            }

                                            pTemp += (datatypes[i].collen+DECIMAL_POINT+SIGN+EXTRA_SPACE)*sizeof(SQLTCHAR);
                                                                                  
                                            break;
                                                                                  
                                        case SQL_TYPE_DATE:
                                                                                  
                                            // remove all leading blanks or tabs
                                            bkptr = wkBuf;
                                            while (bkptr < cptr && *bkptr <= ' ')
                                                bkptr++;
                                            if (cptr != wkBuf && bkptr < cptr)
                                            {
                                                // Converts from YYYYMMDD to MM/DD/YYYY
                                                date = pTemp;
                                                date[0] = bkptr[4];
                                                date[1] = bkptr[5];
                                                date[2] = '/';
                                                date[3] = bkptr[6];
                                                date[4] = bkptr[7];
                                                date[5] = '/';
                                                date[6] = bkptr[0];
                                                date[7] = bkptr[1];
                                                date[8] = bkptr[2];
                                                date[9] = bkptr[3];
                                                date[10] = '\0';
                                                sql_null_data[i] = _tcslen(pTemp)*sizeof(SQLTCHAR);
                                            }

                                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR);

                                            break;
                                                                                  
                                        case SQL_TYPE_TIME:
                                                                          
                                            // remove all leading blanks or tabs
                                            bkptr = wkBuf;
                                            while (bkptr < cptr && *bkptr <= ' ')
                                                bkptr++;
                                                                                  
                                            time = pTemp;
                                            if (cptr != wkBuf && bkptr < cptr)
                                            {
#if !defined(NOSSCANF) //08022001kmt
                                                if (3 == _stscanf(bkptr, DB2e_timeFormat, &x[0], &x[1], &x[2]))
                                                {
                                                    // Converts from "HH.MM.SS" to HH:MM:SS
                                                    time[0] = bkptr[1];
                                                    time[1] = bkptr[2];
                                                    time[2] = ':';
                                                    time[3] = bkptr[4];
                                                    time[4] = bkptr[5];
                                                    time[5] = ':';
                                                    time[6] = bkptr[7];
                                                    time[7] = bkptr[8];
                                                    time[8] = '\0';
                                                    sql_null_data[i] = _tcslen(pTemp)*sizeof(SQLTCHAR);
                                                }
                                                else
                                                {
                                                    print1 (clpstate, _TEXT(CLP_IMPORT_DATA) );
                                                    goto exit;
                                                }
#else//Palm CLP begin
                                                //parsing input to get correct value
                                                k = -1;
                                                d_tag = 1;
                                                memset(astr, 0, ID_LEN);
                                                _tcsncpy((SQLTCHAR*)astr, (SQLTCHAR*)bkptr, 10);
                                                astr[10] = 0;
                                                k = _tcslen((SQLTCHAR*)astr);
                                                d_tag = (Isdigit(astr[1])&&Isdigit(astr[2])
                                                         &&Isdigit(astr[4])&&Isdigit(astr[5])
                                                         &&Isdigit(astr[7])&&Isdigit(astr[8])  
                                                         &&(astr[3] == '.')&&(astr[6] == '.'));  
                                                if ((d_tag)&& (k == 10))
                                                {
                                                    // Converts from "HH.MM.SS" to HH:MM:SS
                                                    time[0] = astr[1];
                                                    time[1] = astr[2];
                                                    time[2] = ':';
                                                    time[3] = astr[4];
                                                    time[4] = astr[5];
                                                    time[5] = ':';
                                                    time[6] = astr[7];
                                                    time[7] = astr[8];
                                                    time[8] = '\0';
                                                    sql_null_data[i] = _tcslen(pTemp)*sizeof(SQLTCHAR);
                                                }
                                                else
                                                {
                                                    print1 (clpstate,(SQLTCHAR*)_TEXT(CLP_IMPORT_DATA) );
                                                    rc = -1;
                                                    goto exit;
                                                }
                                                                                                        
#endif//Palm CLP end    
                                            }
                                                                                  
                                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR);
                                                                                  
                                            break;
                                                                                  
                                        case SQL_TYPE_TIMESTAMP: // d03212001ctl begin
                                                                          
                                            // remove all leading blanks or tabs
                                            bkptr = wkBuf;
                                            while (bkptr < cptr && *bkptr <= ' ')
                                                bkptr++;
                                                                                  
                                            timestamp = pTemp;
                                            if (cptr != wkBuf && bkptr < cptr)
                                            {
                                                // 03232001xh
#if !defined(NOSSCANF) //08022001kmt
                                                numFields = _stscanf(bkptr, DB2e_timestampFormat,
                                                                     &x[0], &x[1], &x[2],
                                                                     &x[3], &x[4], &x[5], &x[6]);
                                                if (7 != numFields)
                                                {
                                                    rc = -1;
                                                    print1 (clpstate, _TEXT(CLP_IMPORT_DATA) );
                                                    goto exit;
                                                }
#else
                                                //parsing input to get correct value
                                                k = -1;
                                                d_tag = 1;
                                                memset(astr, 0, ID_LEN);

                                                _tcsncpy((SQLTCHAR*)astr, (SQLTCHAR*)bkptr, 27);
                                                astr[27] = 0;
                                                k = _tcslen((SQLTCHAR*)astr); //jca
                                                d_tag = (Isdigit(astr[1])&&Isdigit(astr[2])
                                                         &&Isdigit(astr[3])&&Isdigit(astr[4])
                                                         &&Isdigit(astr[6])&&Isdigit(astr[7])
                                                         &&Isdigit(astr[9])&&Isdigit(astr[10])
                                                         &&Isdigit(astr[12])&&Isdigit(astr[13])
                                                         &&Isdigit(astr[15])&&Isdigit(astr[16])
                                                         &&Isdigit(astr[18])&&Isdigit(astr[19])
                                                         &&Isdigit(astr[21])&&Isdigit(astr[22])
                                                         &&Isdigit(astr[23])&&Isdigit(astr[24])
                                                         &&Isdigit(astr[25])&&Isdigit(astr[24])
                                                         &&(astr[5] == '-')&&(astr[8] == '-')
                                                         &&(astr[11] == '-')&&(astr[14] == '.')
                                                         &&(astr[17] == '.')&&(astr[20] == '.')); 
                                                                                                
                                                if (( !d_tag) || (k != 27))
                                                {
                                                    rc = -1;
                                                    print1 (clpstate,(SQLTCHAR*)_TEXT(CLP_IMPORT_DATA) );
                                                    goto exit;
                                                }                  
#endif
                                                memcpy(timestamp, bkptr+1, (datatypes[i].collen)*sizeof(SQLTCHAR));
                                                timestamp[26] = '\0';
                                                sql_null_data[i] = _tcslen(pTemp)*sizeof(SQLTCHAR);
                                            }
                                                                                  
                                            pTemp += (datatypes[i].collen + EXTRA_SPACE)*sizeof(SQLTCHAR);
                                                                                  
                                            break; // d03212001ctl end

                                        default:
                                                                                  
                                            break;
                                                                                  
                                        } // switch

                                        // get the next column
                                        memset(wkBuf, 0 , length * sizeof(SQLTCHAR)); // clean the workspace
                                        cptr = wkBuf;
                                        pBuf++;
                                        i++;
                                        string_read = 0; // 04022001ctl
                                    }
                                } // while

                                // execute the insert statement
                                rc = SQLExecute(hstmt);
                                if (rc)
                                {
                                    ErrorReport(clpstate);
                                    goto exit;
                                }
                                                  
                                nbRec ++; // one more row

                                // done with the entire line; reset and go for the next line
                                pBuf   = readBuf;
                                buflen = 0;
                                memset(pBuf, 0 , length * sizeof(SQLTCHAR));
                                continue;
                            } // if (c != '\n')
                        } // for (idx)
                    } // while (!oFeof)
                } /* if (numcols) */
            }
            else
            {
                print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_IMPORT_CMD) );
                rc = -1;
                goto exit;
            }
        }
        else
        {
            print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_IMPORT_CMD) );
            rc = -1;
            goto exit;
        }
    }
    else
    {
        print1 (clpstate, (SQLTCHAR*)_TEXT(CLP_IMPORT_CMD) );
        rc = -1;
        goto exit;
    }

 exit:

    // print the number of rows that have been successfully inserted
    if (table_name != 0)
    {
        _stprintf((char*)err_string, _TEXT(CLP_IMPORT_OK), nbRec);
        print1 (clpstate, err_string);
    }

    // restore the original state
    if (clpstate->fpIN)
    {
        o_Fclose(clpstate->fpIN);
        clpstate->fpIN = 0;
    }
    clpstate->mode = CLP_INTERACT;

    // free space allocated
    if (sql_null_data)  free(sql_null_data);
    if (datatypes)      free(datatypes);
    if (readBuf)        free(readBuf);
    if (pCSVData)       free(pCSVData);
    if (wkBuf)          free(wkBuf);
    if (int_array)      free(int_array);  // d11182000ctl
    return rc;
}


/********************************************************************************
** DB2eExport()
** We can enhence this routine by parsing the input sentence and ignoring any
** white spaces.
** The output is in CSV format.  Character strings are delimited by a pair
** of double quotes ('"'), and the escape character is also a double quote.
********************************************************************************/
SQLRETURN DB2eExport(CLPState *clpstate)
{
    SQLRETURN     rc = SQL_SUCCESS;
    SQLRETURN     sqlrc;
    SQLTCHAR      exportFormat[] = _TEXT("EXPORT TO %s");
    SQLTCHAR     *exportSQL;
    SQLTCHAR      outPath[ID_LEN];
    SQLTCHAR      outFile[ID_LEN]; 
    SQLTCHAR      *ofPart;
    unsigned int  numFields = 0;
    int           i,j,length;

    // convert the first two words to uppercase
    for (i = 0; i < 9; i++) clpstate->cmd[i] = oToUpper(clpstate->cmd[i]);

//#if defined(NOSSCANF)//Palm CLP begin //08022001kmt
    //parsing input to get correct file name
    _tcscpy((SQLTCHAR*)outPath, (SQLTCHAR*)(clpstate->cmd + 10));
    j = _tcslen((SQLTCHAR*)outPath);
    for(i = 0; i < j; i ++) outPath[i] = oToUpper(outPath[i]); 
    ofPart = (SQLTCHAR*)_tcsstr((char*)outPath, (char*)_TEXT("OF"));
    length = _tcslen((SQLTCHAR*)outPath) - _tcslen((SQLTCHAR*)ofPart) - 1;// -1 if for the ending space
    memset(outFile, 0, ID_LEN);
    _tcsncpy((SQLTCHAR*)outFile, (SQLTCHAR*)(clpstate->cmd + 10), length);
    outFile[length] = 0; //null terminate it
    numFields = 1;
//#else//Palm CLP end
  //  numFields = _stscanf(clpstate->cmd, exportFormat, outFile);
//#endif
  
    if (numFields == 1)
    {
#if defined oEPOC
        _tcscpy(outPath, _TEXT("C:\\System\\Data\\DB2e\\"));
        _tcscat(outPath, outFile);
#elif defined(WIN32U) // 07132001ctl
        memset(outPath, 0, ID_LEN);
        WideCharToMultiByte(CP_ACP,
                            0, 
                            (LPTSTR) outFile,
                            -1,
                            outPath, 
                            ID_LEN,
                            NULL,
                            NULL);
#else
        _tcscpy(outPath, outFile);
#endif
        if ( outPath != 0 )
        {
#if defined(oPALMOS)//Palm CLP begin
            clpstate->fpOUT = o_Fopen((char*)outPath, (char*)_TEXT("wb+"));
            //o_Fopen will only return NULL when there is failure   
            if (clpstate->fpOUT == NULL)
            {
                print2 ( clpstate,(SQLTCHAR*) _TEXT(CLP_EXPORT_FILE), outPath); // 04022001ctl
                rc = -1;
                goto exit;
            }
#else//Palm CLP end
            clpstate->fpOUT = o_Fopen(outPath, _TEXT("wb+")); 
            if (clpstate->fpOUT == 0)
            {
                print2 ( clpstate, _TEXT(CLP_EXPORT_FILE), outPath);
                rc = -1;
                goto exit;
            }
#endif
            // advance the string pointer, passing the export filename
            exportSQL = clpstate->cmd + _tcslen(_TEXT("EXPORT TO ")) + _tcslen(outFile) + 1;

            // convert the following two words to uppercase and compare
            // we support only DEL format
            for (i = 0; i < 7; i++) exportSQL[i] = oToUpper(exportSQL[i]);
            if (0 == _tcsncmp(exportSQL, _TEXT("OF DEL "), 7))
            {
                exportSQL += 7; // advance to the SQL statement for the export
                _tcscpy(clpstate->cmd, exportSQL);
                clpstate->mode = CLP_EXPORT; // switch to EXPORT mode
                sqlrc = CLPExecStatement(clpstate);
            }
            else
            {
                print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_EXPORT_CMD) );
                rc = -1;
                goto exit;
            }
        }
        else
        {
            print2 ( clpstate, (SQLTCHAR*)_TEXT(CLP_EXPORT_FILE), outPath);
            rc = -1;
            goto exit;
        }
    }
    else
    {
        rc = -1;
        print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_EXPORT_CMD) );
        goto exit;
    }

 exit:
    // restore the original state
    if (clpstate->fpOUT)
    {
        o_Fclose(clpstate->fpOUT);
        clpstate->fpOUT = 0;
    }
    clpstate->mode = CLP_INTERACT;
    return rc;
}


/********************************************************************************
** CLPDescribeStatement()
**   This function performs a "DESCRIBE" of a SQL statement, returning
**   data types, length, column names and nullability information.
********************************************************************************/
SQLRETURN CLPDescribeStatement(CLPState *clpstate)
{
    SQLHSTMT hstmt = clpstate->hstmt;
    SQLTCHAR *sqlstmt = &(clpstate->cmd[0]);
    SQLRETURN rc = SQL_SUCCESS;
    SQLSMALLINT numCols = 0;
    unsigned int rowCount = 0;
    SQLTCHAR string[300];
    SQLTCHAR *s;
    int i;
    SQLTCHAR colname[32];
    SQLSMALLINT coltype, colnamelen, nullable;
    SQLUINTEGER collen;
    SQLUINTEGER j;
    SQLSMALLINT scale;

    //  Check to see if we need to set statement attributes
    rc = CLPSetStmtMode(clpstate);
    
    //  Executes the statement
    sqlstmt += 9;  // skip "describe"
    if (SQL_SUCCESS == rc || SQL_SUCCESS_WITH_INFO == rc)
        rc = SQLExecDirect(hstmt, sqlstmt, SQL_NTS); 
    if (rc)
    {
        writeOut(clpstate, (SQLTCHAR *)_TEXT("Invalid SELECT statement to be described.\n"));
        ErrorReport(clpstate);
        return rc;
    }

    rc = SQLNumResultCols(hstmt, &numCols);
    if (rc)
    {
        writeOut(clpstate, (SQLTCHAR *)_TEXT("Invalid SELECT statement to be described.\n"));
        ErrorReport(clpstate);
        return rc;
    }
            
    if (numCols)               // Assume that SELECT statement has been executed
    {  
        s = string;
        _tcscpy(s, _TEXT("sqltype               "));
        s += _tcslen(_TEXT("sqltype               "));

        _tcscpy(s, _TEXT("sqllen  "));
        s += _tcslen(_TEXT("sqllen  "));

        _tcscpy(s, _TEXT("column name                     "));
        s += _tcslen(_TEXT("column name                     "));

        _tcscpy(s, _TEXT("nullability"));
        s += _tcslen(_TEXT("nullability"));

        *s++ = 0;


        writeOut(clpstate, (SQLTCHAR *)string);

        s = string;
        _tcscpy(s, _TEXT(" --------------------  ------  ------------------------------  -----------"));
        writeOut(clpstate, (SQLTCHAR *)string);

        for (i = 0; i < numCols; i++) 
        {
            s = string;
            *s++ = ' ';
            SQLDescribeCol(hstmt, ( SQLSMALLINT ) (i + 1),
                           colname,
                           sizeof(colname),
                           (short *)&colnamelen,
                           (short *)&coltype,
                           &collen,
                           (short *)&scale,
                           (short *)&nullable);
            switch (coltype)
            {
            case SQL_CHAR:
                _tcscpy(s, _TEXT("CHARACTER             "));
                s += _tcslen(_TEXT("CHARACTER             "));

                break;

            case SQL_VARCHAR:
                _tcscpy(s, _TEXT("VARCHAR               "));
                s += _tcslen(_TEXT("VARCHAR               "));

                break;

            case SQL_TYPE_DATE:
                _tcscpy(s, _TEXT("DATE                  "));
                s += _tcslen(_TEXT("DATE                  "));
                collen = 10; // fixed length datatype
                    
                break;

            case SQL_TYPE_TIME:
                _tcscpy(s, _TEXT("TIME                  "));
                s += _tcslen(_TEXT("TIME                  "));
                collen = 8; // fixed length datatype

                break;

            case SQL_DECIMAL:
                _tcscpy(s, _TEXT("DECIMAL               "));
                s += _tcslen(_TEXT("DECIMAL               "));

                break;

            case SQL_INTEGER:
                _tcscpy(s, _TEXT("INTEGER               "));
                s += _tcslen(_TEXT("INTEGER               "));
                collen = 4; // fixed length datatype

                break;

            case SQL_SMALLINT:
                _tcscpy(s, _TEXT("SMALLINT              "));
                s += _tcslen(_TEXT("SMALLINT              "));
                collen = 2; // fixed length datatype

                break;

            case SQL_BLOB:
                _tcscpy(s, _TEXT("BLOB                  "));
                s += _tcslen(_TEXT("BLOB                  "));

                break;
                          
            case SQL_TYPE_TIMESTAMP:
                _tcscpy(s, _TEXT("TIMESTAMP             "));
                s += _tcslen(_TEXT("TIMESTAMP             "));

                break;

            default:

                break;
            }

#if defined(oEPOC) && defined(UNICODE) //08022001kmt
            if (coltype == SQL_DECIMAL) // 3114 thf
                _stprintf((char *)s, _TEXT("%3d,%2d"), (long) collen, (long) scale); // 3690 thf
            else
                _stprintf((char *)s, _TEXT("%6d"), (long) collen); // 3690 thf
#else
            if (coltype == SQL_DECIMAL) // 3114 thf
                _stprintf((char *)s, _TEXT("%3ld,%2ld"), (long) collen, (long) scale); // 3690 thf
            else
                _stprintf((char *)s, _TEXT("%6ld"), (long) collen); // 3690 thf
#endif
               
            s += 6;

            _tcscpy(s, _TEXT("  "));
            s += 2;

            _tcsncpy(s, colname, colnamelen);
            s += colnamelen;
            for (j = colnamelen; j < 18; j++) *s++ = ' ';
            *s++ = ' ';

            _tcscpy(s, _TEXT("             "));
            s += _tcslen(_TEXT("             "));
            if (nullable) {_tcscpy(s, _TEXT("NULL")); s += 4;}
            else {_tcscpy(s, _TEXT("NOT NULL")); s += 8;}
            *s++ = 0;
            writeOut(clpstate, (SQLTCHAR *)string);
        }
          
    }
    else
    {
        writeOut(clpstate, (SQLTCHAR *)_TEXT("An unexpected token was found following \"DESCRIBE\".\n"));
        writeOut(clpstate, (SQLTCHAR *)_TEXT("Expected token may include: \"SELECTT\".\n"));
    }
    
    return rc;
}

// 04022001ctl begin
/********************************************************************************
 ** processTableCommands()
 **   This function takes a table of 2 columns:
 **     first column (ID) is an integer containing the sequence number of execution;
 **     second column (SQL) is a character column containing a SQL statement to be executed;
 **   It reads every row from the table in the increasing order of the first column and executes each statement.
 **   Optionally, it outputs the result into a table specified by outTable.
 **   Here is an example.
 **       CREATE TABLE SQLSTMT (ID INT PRIMARY KEY, TXT VARCHAR(2048))
 **       CREATE TABLE RESULT  (ID INT PRIMARY KEY, TXT VARCHAR(2048))
 **       -- insert or import SQL statements to SQLSTMT table
 **       -- execute those SQL statements and store the result in RESULT table:
 **       $table SQLSTMT RESULT
 **
********************************************************************************/
int processTableCommands(CLPState *clpstate, SQLTCHAR *inTable, SQLTCHAR *outTable)
{
    SQLHSTMT      hstmt = clpstate->hstmt;
    SQLTCHAR     *sqlstart = (&(clpstate->cmd[0]));
    CLPState     *my_clpstate = 0;
    SQLHSTMT          my_hstmt = -1;
    SQLHSTMT          outhstmt;
    SQLRETURN     rc;
    DataTypes     datatypes[2];
    long          str1len;
    SQLSMALLINT   numCols = 0;
    SQLTCHAR      tableCMDsql[BUFSIZE];  // for constructing "SELECT max(ID) FROM outTable"
    DataTypes     coldatatypes;

    // Open the output table, if specified, for storing the result.
    // Note that we use clpstate here instead of my_clpstate!
    // create a new statement handle -- outhstmt.
    if (outTable)
    {
        rc = SQLAllocHandle(SQL_HANDLE_STMT, clpstate->hdbc, &(clpstate->outhstmt));
        if (SQL_SUCCESS != rc)
        { // if allocation fails
            goto exit;
        }

        outhstmt = clpstate->outhstmt;
        // Prepare the statement: "SELECT * FROM outTable"
        tableCMDsql[0] = '\0';
        _tcscpy(tableCMDsql, _TEXT("SELECT * FROM "));
        _tcscat(tableCMDsql, outTable);
          
        // Executes the statement: "SELECT * FROM outTable"
        rc = SQLExecDirect(outhstmt, tableCMDsql, SQL_NTS);
        if (rc)
        {
            ErrorReport(clpstate);
            goto exit;
        }

        rc = SQLNumResultCols(outhstmt, &numCols);
        if (rc)
        {
            ErrorReport(clpstate);
            goto exit;
        }

        if (numCols != 2)
        {
            print1 (clpstate, (SQLTCHAR*)_TEXT(CLP_TABLE_CMD_ERROR));
            goto exit;
        }

        // make sure the first column is an integer column
        SQLDescribeCol(outhstmt, ( SQLSMALLINT ) 1,
                       coldatatypes.colname,
                       sizeof(coldatatypes.colname),
                       (short *)&(coldatatypes.colnamelen),
                       (short *)&(coldatatypes.coltype),
                       &(coldatatypes.collen),
                       (short *)&(coldatatypes.scale),
                       (short *)&(coldatatypes.nullable));
        if (coldatatypes.coltype != SQL_INTEGER)
        {
            print1 (clpstate, (SQLTCHAR*)_TEXT(CLP_TABLE_CMD_ERROR));
            goto exit;
        }

        // make sure the second column is a character column
        SQLDescribeCol(outhstmt, ( SQLSMALLINT ) 2,
                       coldatatypes.colname,
                       sizeof(coldatatypes.colname),
                       (short *)&(coldatatypes.colnamelen),
                       (short *)&(coldatatypes.coltype),
                       &(coldatatypes.collen),
                       (short *)&(coldatatypes.scale),
                       (short *)&(coldatatypes.nullable));
        if (coldatatypes.coltype != SQL_VARCHAR && coldatatypes.coltype != SQL_CHAR)
        {
            print1 (clpstate, (SQLTCHAR*)_TEXT(CLP_TABLE_CMD_ERROR));
            goto exit;
        }

        // store the column length
        clpstate->outTableTextLen = coldatatypes.collen;
        _tcscpy(clpstate->outTable, outTable);

        // Prepare the statement: "SELECT MAX(ID) FROM outTable"
        tableCMDsql[0] = '\0';
        _tcscpy(tableCMDsql, (SQLTCHAR*)_TEXT("SELECT MAX(ID) FROM "));
        _tcscat(tableCMDsql, outTable);
          
        // Executes the statement: "SELECT MAX(ID) FROM outTable"
        rc = SQLExecDirect(clpstate->outhstmt, tableCMDsql, SQL_NTS);
        if (rc)
        {
            ErrorReport(clpstate);
            goto exit;
        }
        // Need to get the maximum sequence value in the outTable.
        if (SQL_SUCCESS == (rc = SQLFetch(clpstate->outhstmt)))
        {
            rc = SQLGetData(clpstate->outhstmt, 1, SQL_C_LONG, &(clpstate->outTablecount),
                            sizeof(clpstate->outTablecount), (long *) &str1len);
            if (rc) // 07232001ctl
            {
                ErrorReport(clpstate);
                goto exit;
            }
        }
        SETBIT(clpstate->mode, CLP_TABLE_MODE);
    }
          
    // Allocate CLPState structure for reading the table of SQL statements
    // We use my_clpstate and my_hstmt for reading SQL statements to be executed, and
    // use clpstate and outhstmt to store the results of these SQL execution.
    my_clpstate = (CLPState *) malloc(sizeof(CLPState));
    if (!my_clpstate)
    {
        print1 ( clpstate, (SQLTCHAR*)_TEXT(CLP_INTERNAL_ERROR2) );
        rc = -1;
        goto exit;
    }
    my_clpstate->mode     = CLP_INTERACT;
    my_clpstate->fpIN     = 0;
    my_clpstate->fpOUT    = 0;
    my_clpstate->henv     = clpstate->henv; // use the same environment handle
    my_clpstate->hdbc     = clpstate->hdbc; // use the same connect handle
    my_clpstate->hstmt    = my_hstmt;
    my_clpstate->stmtCtrl = DB2eCLP_STMTATTR_UNCHANGED;

    // create a new statement handle
    rc = SQLAllocHandle(SQL_HANDLE_STMT, my_clpstate->hdbc, &my_hstmt);
    if (SQL_SUCCESS != rc)
    { // if allocation fails
        goto exit;
    }       
    else
        my_clpstate->hstmt = my_hstmt;
    _tcscpy(my_clpstate->cmd, inTable);

    //  Executes the statement: "SELECT ID, TXT FROM inTable ORDER BY ID"
    rc = SQLExecDirect(my_hstmt, my_clpstate->cmd, SQL_NTS);
    if (rc)
    {
        ErrorReport(my_clpstate);
        goto exit;
    }

    rc = SQLNumResultCols(my_hstmt, &numCols);
    if (rc)
    {
        ErrorReport(my_clpstate);
        goto exit;
    }
            
    if (numCols == 2)             // must have only 2 columns
    {
        // make sure that the data type for the first column is integer/small
        SQLDescribeCol(my_hstmt, ( SQLSMALLINT ) 1,
                       datatypes[0].colname,
                       sizeof(datatypes[0].colname),
                       (short *)&(datatypes[0].colnamelen),
                       (short *)&(datatypes[0].coltype),
                       &(datatypes[0].collen),
                       (short *)&(datatypes[0].scale),
                       (short *)&(datatypes[0].nullable));
        if (!(datatypes[0].coltype == SQL_INTEGER || datatypes[0].coltype == SQL_SMALLINT)) {
            print1(my_clpstate, (SQLTCHAR*)_TEXT(CLP_TABLE_CMD_ERROR2));
        }

        // make sure that the data type for the second column is varchar/char
        SQLDescribeCol(my_hstmt, ( SQLSMALLINT ) 2,
                       datatypes[1].colname,
                       sizeof(datatypes[1].colname),
                       (short *)&(datatypes[1].colnamelen),
                       (short *)&(datatypes[1].coltype),
                       &(datatypes[1].collen),
                       (short *)&(datatypes[1].scale),
                       (short *)&(datatypes[1].nullable));
        if (!(datatypes[1].coltype == SQL_CHAR || datatypes[1].coltype == SQL_VARCHAR)) {
            print1(my_clpstate, (SQLTCHAR*)_TEXT(CLP_TABLE_CMD_ERROR2));
        }

        // Loop through the enable result table, one row at a time and output the data
        while (SQL_SUCCESS == (rc = SQLFetch(my_hstmt)))
        {
            memset(sqlstart, 0, MAX_CMD_LEN);
            rc = SQLGetData(my_hstmt, 2 /* second column */, SQL_C_CHAR, sqlstart, MAX_CMD_LEN, &str1len);
            if (rc) // 07232001ctl
            {
                ErrorReport(clpstate);
                goto exit;
            }
            if (str1len > 0) 
            {
                writeOut(clpstate, sqlstart);

                // execute the SQL statement which is just retrieved from the table
                CLPcommand_interpreter(clpstate);
            }
            else
                continue; // a null row
        }
    }

 exit:
    if (my_clpstate != 0)
        free (my_clpstate);

    // free the statement handle
    if (my_hstmt != -1)
        SQLFreeHandle(SQL_HANDLE_STMT, my_hstmt);
        
    CLRBIT(clpstate->mode, CLP_TABLE_MODE);

    return rc;
}
// 04022001ctl end

/********************************************************************************
** processBatch()
**   This function takes a file of SQL statements and executes the statements.
**   It outputs the result into the output file.
********************************************************************************/
int processBatch(CLPState *clpstate, SQLTCHAR *inFile, SQLTCHAR *outFile)
{
#if !defined(oEPOC)  && !defined(oWINCE) && !(defined oPALMOS)
    SQLHSTMT hstmt = clpstate->hstmt;
    SQLTCHAR *sqlstart = (&(clpstate->cmd[0]));
    SQLRETURN sqlrc = 0;
    SQLTCHAR *filebuf = 0;
    SQLTCHAR *nextAddr = 0;
    SQLTCHAR *bufpos  = 0;
    SQLTCHAR *bufendAddr = 0;
    SQLTCHAR *sqlstmt = 0;
    struct stat fstatus;
    SQLTCHAR *pBuf = 0;
    int newStmt = 1;
    int inString = 0;
    int buflen = 0;
    SQLTCHAR c = 0;
   

    // Read the input file
    if (0 == stat(inFile, &fstatus))
    {
        filebuf = malloc(fstatus.st_size);
                    
        if (filebuf)
        {
            clpstate->fpIN = o_Fopen(inFile, _TEXT("rb"));
                                           
            if (clpstate->fpIN)
            {
                o_Fread(filebuf, 1, fstatus.st_size, clpstate->fpIN);
                bufpos = filebuf;
                bufendAddr = filebuf + fstatus.st_size;
                o_Fclose(clpstate->fpIN);

                if ( outFile != 0 )
                {
                    clpstate->fpOUT = o_Fopen(outFile, _TEXT("wb+")); 
                                                           
                    if (0 == clpstate->fpOUT)
                    {
                        print2(clpstate, _TEXT("Unable to open file %s \n"), outFile);
                        o_Fflush(stdout);
                        free(filebuf);
                        return 0;
                    }
                }
                else
                    clpstate->fpOUT = 0;

                // process the batch file
                nextAddr = bufpos;
                while (nextAddr < bufendAddr)
                {
                    c = *nextAddr++;

                    // check to see if it exceeds the maximum buffer length
                    // MAX_CMD_LEN characters allow only for each statement.
                    if (++buflen >= MAX_CMD_LEN)
                    {
                        print1(clpstate, _TEXT("   ERROR:  Command exceeded the maximum allowable length.  \n"));
                        o_Fflush(stdout);
                        o_Fclose(clpstate->fpOUT);
                        free(filebuf);
                        return -1;
                    }

                    if ( c != '\n' && newStmt )
                    {
                        pBuf = sqlstart;
                        memset(pBuf, 0, MAX_CMD_LEN);
                        newStmt = 0;
                    }

                    if ('\'' == c)
                    {
                        if (inString)
                            inString = 0;
                        else
                            inString = 1;
                        *pBuf++ = c;
                        continue;
                    }

                    // anything in between quotes
                    if (inString)
                    {
                        if (c != '\n' && c != 0x0D)
                            *pBuf++ = c;
                        continue;
                    }

                    if (';' == c )
                    {
                        if (inString)
                            *pBuf++ = c;
                        else   // stmt completes
                        {
                            *pBuf = 0;
                            // sqlrc = DB2eExecStatement(clpstate);
                            CLPcommand_interpreter(clpstate);
                            newStmt = 1;
                            inString = 0;
                            buflen = 0;
                        }
                        continue;
                    }

                    //  comment
                    if ('#' == c)
                    {
                        while (nextAddr < bufendAddr)
                        {
                            c = *nextAddr++;
                            if (0x0A == c)
                                break;
                        }
                        continue;
                    }

                    if ('-' == c)
                    {
                        if ('-' == *nextAddr)     // Skip commentary line
                        {
                            while (nextAddr < bufendAddr)
                            {
                                c = *nextAddr++;
                                if (0x0D == c || 0x0A == c) {
                                    break;
                                }
                            }
                            continue;
                        }
                    }

                    // ignore carriage return CR and line feed LF
                    if (0x0D == c || 0x0A == c)
                        continue;

                    *pBuf++ = c;
                }
            }
            else {
                print2(clpstate, _TEXT("Unable to open %s \n"), inFile);
                o_Fflush(stdout);
            }
        }
        else {
            print1(clpstate, _TEXT("Unable to allocate memory \n"));
            o_Fflush(stdout);
        }
    }
    else {
        print2 (clpstate, _TEXT("Unable to access the file %s \n"), inFile);
        o_Fflush(stdout);
    }


    if (filebuf)
        free(filebuf);
   
    if (clpstate->fpOUT)
    {
        o_Fclose(clpstate->fpOUT);
        clpstate->fpOUT = 0;
    }
#endif
    return 0;
}


/********************************************************************************
** CLPcommand_interpreter()
**   This function interpretes the parsed command line
**   and calls its associated function to complete the
**   request.
********************************************************************************/
int CLPcommand_interpreter(CLPState *clpstate)
{
    SQLHSTMT  hstmt = clpstate->hstmt;
    SQLRETURN sqlrc = 0;
    SQLTCHAR *cmdstart = &(clpstate->cmd[0]);
    SQLTCHAR  buf[BUFSIZE], *cmdbuf; //08022001kmt
    SQLTCHAR  fileFormat[] = _TEXT("$FILE %s %s");
    SQLTCHAR  inPath[BUFSIZE];
    SQLTCHAR  outPath[BUFSIZE];

    SQLTCHAR  inFile[BUFSIZE]; 
    SQLTCHAR  outFile[BUFSIZE];
    int       i, j, len;
    int       rc = 0;
    SQLTCHAR  chmodFormat[] = _TEXT("$CHMOD %s");
    SQLTCHAR  mode[10];
    SQLTCHAR  *pstr, *bstr;
    int       imode = DB2eCLP_STMTATTR_UNCHANGED;
    unsigned int  numFields = 0;

    // d4254qz
    SQLTCHAR  createstmthdFormat[] = _TEXT("CREATE STMT HANDLE :%s");
    SQLTCHAR  mode1[10];
	
    SQLTCHAR  dropstmthdFormat[] = _TEXT("DROP STMT HANDLE :%s");
    SQLTCHAR  mode2[10];
    
    SQLTCHAR  execdirectstmthdFormat[] = _TEXT("EXECDIRECT :%s %s");
    SQLTCHAR  mode3[10];
    SQLTCHAR  mode10[PARTSIZE];  // assume max stmt length is PARTSIZE?
    SQLTCHAR  *stmtpart;
    
    SQLTCHAR  fetchstmthdFormat[] = _TEXT("FETCH :%s");
    SQLTCHAR  mode4[10];
    SQLINTEGER str1len;
    SQLSMALLINT numcolsfetched = 0;
    SQLTCHAR * resultrow = 0;
    SQLTCHAR * onecol = 0;

    SQLTCHAR  preparestmthdFormat[] = _TEXT("PREPARE :%s %s");
    SQLTCHAR  mode5[10];
    SQLTCHAR  mode11[PARTSIZE];   // assume the max length of the first string in the stmt is PARTSIZE ? 

    SQLTCHAR  executestmthdFormat[] = _TEXT("EXECUTE :%s");
    SQLTCHAR  mode6[10];
    
    SQLTCHAR  bindstmthdFormat[] = _TEXT("BIND :%s %s %s");
    SQLTCHAR  mode7[10];
    SQLTCHAR  mode8[10];
    SQLTCHAR  mode9[PARTSIZE];  // assume the max length of the first string in the stmt is PARTSIZE ? 
    int param_no =0;
    SQLSMALLINT    numcols = 10;  // assume 10 at this time 
    SQLTCHAR **paramvalue = 0; // d4254qz

    // 04022001ctl begin
    SQLTCHAR  tableFormat[] = _TEXT("$TABLE %s %s");
    SQLTCHAR  inTable[ID_LEN];      // input tablename in $TABLE command
    SQLTCHAR  outTable[ID_LEN];     // output tablename in $TABLE command
    SQLTCHAR  tableCMDsql[BUFSIZE];  // for constructing "SELECT ID, TXT FROM inTable"
    // 04022001ctl end

    SQLSMALLINT templen;

    clpstate->flag_hstmt = 1; //xh 092801

    if (!(cmdbuf = malloc(MAX_CMD_LEN))) //08022001kmt
      goto exit;

    // ignore leading white spaces
    while (' ' == *cmdstart || '\t' == *cmdstart || '\n' == *cmdstart)
        cmdstart++;

    //added by 03192001xh to remove the leading space of a command
    memset(cmdbuf, 0, MAX_CMD_LEN);
    _tcscpy(cmdbuf, cmdstart);
    memset(clpstate->cmd, 0, MAX_CMD_LEN);
    _tcscpy(clpstate->cmd, cmdbuf);
  
    //d3662: reset it to correct value xh06212001
    cmdstart = &(clpstate->cmd[0]);

    // copy to the local buffer and convert the string to uppoercase
    _tcsncpy(buf, cmdbuf, BUFSIZE);
    //_tcsncpy(buf, cmdstart, BUFSIZE); !!! doesn't work, as cmdstart point to wrong
    //                                  !!! place after _tcsncpy
    for (i = 0; i < BUFSIZE; i++) buf[i] = oToUpper(buf[i]);

    if (0 == _tcsncmp(buf, _TEXT("QUIT"), 4) && TSTBIT(clpstate->mode,CLP_INTERACT))
      { 
      // d4254qz
      for(j =0; j < 20; j++)
        {
        if (clpstate->bindarray[j] != NULL)
          {
          paramvalue = clpstate->bindarray[j];
          if(paramvalue){
           for(i=0; i<numcols; i++)
            {
            if (paramvalue[i]){
              free(paramvalue[i]);
              paramvalue[i] = NULL;
              }
            }
           free(paramvalue);
           clpstate->bindarray[imode] = NULL;
           //paramvalue = NULL;
           }
          }
        }
      
      free(cmdbuf);
      return CLP_QUIT;
      
      }
//#if !defined(NOSSCANF)    
    //d4254qz
    else if (0 == _tcsncmp(buf, _TEXT("CREATE STMT HANDLE "), 19)) 
      {
        clpstate->flag_hstmt = 0; //xh 092801
      if (clpstate->mstmtno > 20) 
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("current statement handles already reach the maximum number"));
        goto exit;
        }
#if !defined(NOSSCANF)      
      numFields = _stscanf(buf, createstmthdFormat, mode1);
#else
      numFields = oSscanf(buf,(SQLTCHAR*) _TEXT("CREATE STMT HANDLE"), mode1, NULL, NULL);
#endif
      if (1 == numFields)
        {
        imode = atoi(mode1);
        }
      else 
        {
        imode = -1;
        goto exit;
        }
      
      if(imode < 0 || imode > 19)
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("the statement handle number is out of range (0~19) "));
        goto exit;
        }
      
      if(clpstate->mhstmt[imode] == 0)        
        {
        sqlrc=SQLAllocHandle(SQL_HANDLE_STMT, clpstate->hdbc, &clpstate->mhstmt[imode]);
        clpstate->current_mhstmt = imode;
        clpstate->mstmtno +=1;      
        }
      else
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("This statement handle already exists"));
        goto exit;
        }
      if (!sqlrc) 
         writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
       else
         {
        ErrorReport(clpstate);
        goto exit;
         }
      }
    else if (0 == _tcsncmp(buf, _TEXT("DROP STMT HANDLE "), 17)) 
      {
      clpstate->flag_hstmt = 0; //xh 092801
#if !defined(NOSSCANF)      
      numFields = _stscanf(buf, dropstmthdFormat, mode2);
#else
      numFields = oSscanf(buf, (SQLTCHAR*)_TEXT("DROP STMT HANDLE"), mode2, NULL, NULL);
#endif
      if (1 == numFields)
        {
        imode = atoi(mode2);
        }
      else 
        {
        imode = -1;
        goto exit;
        }
      
      if(imode < 0 || imode > 19)
        {
	       print1 (clpstate, (SQLTCHAR*)_TEXT("the statement handle number is out of range (0~19) "));
         goto exit;
        }
      
      if(clpstate->mhstmt[imode] != 0) 
        
        { 
        if (clpstate->bindarray[imode] != NULL)
          {
          paramvalue = clpstate->bindarray[imode];
          if(paramvalue){
           for(i=0; i<numcols; i++)
            {
              if(paramvalue[i]){
                 free(paramvalue[i]);
                 paramvalue[i] = NULL;
                }
            }
           free(paramvalue);
           clpstate->bindarray[imode] = NULL;
           //paramvalue = NULL;
           }
          }
        
        sqlrc = SQLFreeHandle(SQL_HANDLE_STMT, clpstate->mhstmt[imode]);
        
        clpstate->mstmtno -=1;
        clpstate->mhstmt[imode] = 0;
        
        }
      else
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("No such statement handle exists"));
        goto exit;
        }
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      else
         {
        ErrorReport(clpstate);
        goto exit;
         }
      }
    else if (0 == _tcsncmp(buf, _TEXT("EXECDIRECT "), 11)) 
      {
        clpstate->flag_hstmt = 0; //xh 092801
      //numFields = _stscanf(buf, execdirectstmthdFormat, mode3, mode10);
		    for (i = 0; i < 11; i++) cmdbuf[i] = oToUpper(cmdbuf[i]);    
#if !defined(NOSSCANF)      
      numFields = _stscanf(cmdbuf, execdirectstmthdFormat, mode3, mode10);
#else
      numFields = oSscanf(cmdbuf, (SQLTCHAR*)_TEXT("EXECDIRECT"), mode3, mode10, NULL);
#endif
        
        if (2 == numFields)
          {
          imode = atoi(mode3);
          stmtpart = (SQLTCHAR*)_tcsstr(cmdbuf, mode10);
          
          }
        else 
          {
          imode = -1;
          goto exit;
          }
        
        if(imode < 0 || imode > 19)
          {
          print1 (clpstate, (SQLTCHAR*)_TEXT("the statement handle number is out of range (0~19) "));
          goto exit;
          }else
          clpstate->current_mhstmt = imode; //d4508 xh
        
        if(clpstate->mhstmt[imode] == 0) 
          
          {
          print1 (clpstate, (SQLTCHAR*)_TEXT("No such statement handle exists"));
          goto exit;
          }
        else
          {
          if (clpstate->bindarray[imode] != NULL)
            {
            paramvalue = clpstate->bindarray[imode];
            if(paramvalue){
             for(i=0; i<numcols; i++)
              {
              if (paramvalue[i]){
                free(paramvalue[i]);
                paramvalue[i] = NULL;
                }
              }
             free(paramvalue);
             clpstate->bindarray[imode] = NULL;
             //paramvalue = NULL;
             }
           }
          
          sqlrc = SQLExecDirect(clpstate->mhstmt[imode], stmtpart, SQL_NTS);
          
          }
        
        if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
        else
         {
        ErrorReport(clpstate);
        goto exit;
         }
      }
    else if (0 == _tcsncmp(buf, _TEXT("FETCH "), 6)) 
      {
      clpstate->flag_hstmt = 0; //xh 092801
#if !defined(NOSSCANF)      
      numFields = _stscanf(buf, fetchstmthdFormat, mode4);
#else
      numFields = oSscanf(buf, (SQLTCHAR*)_TEXT("FETCH"), mode4, NULL, NULL);
#endif
      if (1 == numFields)
        {
        imode = atoi(mode4);
        }
      else 
        {
        imode = -1;
        goto exit;
        }
      
      if(imode < 0 || imode > 19)
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("the statement handle number is out of range (0~19) "));
        goto exit;
        }else
       clpstate->current_mhstmt = imode; //d4508 xh
      
      if(clpstate->mhstmt[imode] == 0) 
        
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("No such statement handle exists"));
        goto exit;
        }
      else
        {
        sqlrc=SQLFetch(clpstate->mhstmt[imode]);
        if (sqlrc == SQL_SUCCESS)
          {
          sqlrc = SQLNumResultCols(clpstate->mhstmt[imode], &numcolsfetched);
          if (sqlrc)
            {
            ErrorReport(clpstate);
            goto exit;
            }
          
          if (numcolsfetched)
            {
            onecol = (SQLTCHAR *)malloc(BUFSIZE);  //assume the column size is BUFSIZE
            memset(onecol, 0, BUFSIZE);
            
            resultrow= (SQLTCHAR *)malloc(numcolsfetched * (BUFSIZE+ sizeof(SQLTCHAR)/* for extra space */)); 
            memset(resultrow, 0, numcolsfetched * (BUFSIZE+ sizeof(SQLTCHAR)));
            
            for(i=1; i<=numcolsfetched; i++)
              {
              sqlrc = SQLGetData(clpstate->mhstmt[imode], i, SQL_C_CHAR, onecol, MAX_CMD_LEN, &str1len);
              if (sqlrc) 
                {
                ErrorReport(clpstate);
                goto exit;
                }
              if (str1len > 0)
                _tcscat(resultrow, onecol); 
              _tcscat(resultrow, " "); 
              }
            
            writeOut(clpstate, resultrow);
            
            free(onecol);
            free(resultrow);
            }
          }   
        }
      
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      else if(sqlrc != 100) //come to the end of result set
         {
        ErrorReport(clpstate);
        goto exit;
         }
      }
    else if (0 == _tcsncmp(buf, _TEXT("PREPARE "), 8)) 
      {
      
      clpstate->flag_hstmt = 0; //xh 092801
      //numFields = _stscanf(buf, preparestmthdFormat, mode5, mode11);
      for (i = 0; i < 8; i++) cmdbuf[i] = oToUpper(cmdbuf[i]); 
#if !defined(NOSSCANF)      
      numFields = _stscanf(cmdbuf, preparestmthdFormat, mode5, mode11);
#else
      numFields = oSscanf(cmdbuf, (SQLTCHAR*)_TEXT("PREPARE"), mode5, mode11, NULL);
#endif     
      if (2 == numFields)
        {
        imode = atoi(mode5);
        stmtpart = (SQLTCHAR*)_tcsstr(cmdbuf, mode11);
        }
      else 
        {
        imode = -1;
        goto exit;
        }
      
      if(imode < 0 || imode > 19)
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("the statement handle number is out of range (0~19) "));
        goto exit;
        }else
       clpstate->current_mhstmt = imode; //d4508 xh
      
      if(clpstate->mhstmt[imode] == 0) 
        
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("No such statement handle exists"));
        goto exit;
        }
      else
        {
        if (clpstate->bindarray[imode] != NULL)
          {
          paramvalue = clpstate->bindarray[imode];
          if(paramvalue){
           for(i=0; i< numcols; i++)
            {
            if (paramvalue[i]){
              free(paramvalue[i]);
              paramvalue[i] = NULL;
              }
            }
           free(paramvalue);
           clpstate->bindarray[imode] = NULL;
           //paramvalue = NULL;
           }
          }
        sqlrc=SQLPrepare(clpstate->mhstmt[imode], stmtpart, SQL_NTS);
        }
      
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      else
         {
        ErrorReport(clpstate);
        goto exit;
         }
      }
    else if (0 == _tcsncmp(buf, _TEXT("BIND "), 5)) 
      {
      clpstate->flag_hstmt = 0; //xh 092801
      //numFields = _stscanf(buf, bindstmthdFormat, mode7, mode8, mode9);
      for (i = 0; i < 5; i++) cmdbuf[i] = oToUpper(cmdbuf[i]);
#if !defined(NOSSCANF)      
      numFields = _stscanf(cmdbuf, bindstmthdFormat, mode7, mode8, mode9);
#else
      numFields = oSscanf(cmdbuf, (SQLTCHAR*)_TEXT("BIND"), mode7, mode8, mode9);
#endif      
      if (3 == numFields)
        {
        imode = atoi(mode7);
        param_no = atoi(mode8); 
        }
      else 
        {
        imode = -1;
        goto exit;
        }
      
      if(imode < 0 || imode > 19)
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("the statement handle number is out of range (0~19) "));
        goto exit;
        }else
       clpstate->current_mhstmt = imode; //d4508 xh
      
      if(clpstate->mhstmt[imode] == 0) 
        
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("No such statement handle exists"));
        goto exit;
        }
      else
        {     
        if (clpstate->bindarray[imode] == NULL) 
          {
          paramvalue = malloc(sizeof(SQLTCHAR *) * numcols );
          
          for(i=0; i<numcols; i++)
            {
            paramvalue[i] = (SQLTCHAR*) malloc(sizeof(SQLTCHAR) * BUFSIZE); // use BUFSIZE for the max length of input string 
            
            }
          clpstate->bindarray[imode] = paramvalue;
          
          }
        else
          paramvalue = clpstate->bindarray[imode]; 
        
        _tcscpy(paramvalue[param_no-1], mode9);
        
        // d4254qz only test SQL_CHAR at this time
        /* sqlrc=SQLBindParameter(clpstate->mhstmt[imode], 
        param_no, 
        SQL_PARAM_INPUT, 
        SQL_C_CHAR, 
        SQL_CHAR, 
        _tcslen(paramvalue[param_no-1]) * sizeof(SQLTCHAR), 
        0, 
        paramvalue[param_no-1], 
        _tcslen(paramvalue[param_no-1]) * sizeof(SQLTCHAR), 
        NULL);
        */
        sqlrc=SQLBindParameter(clpstate->mhstmt[imode], 
                    param_no, 
										SQL_PARAM_INPUT, 
                    SQL_C_CHAR, 
                    SQL_CHAR, 
                    0, 
                    0, 
                    paramvalue[param_no-1], 
                    BUFSIZE * sizeof(SQLTCHAR), 
                    NULL);
        
        } 
      
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      else
         {
        ErrorReport(clpstate);
        goto exit;
         }
      } 
    else if (0 == _tcsncmp(buf, _TEXT("EXECUTE "), 8)) 
      {
      clpstate->flag_hstmt = 0; //xh 092801
#if !defined(NOSSCANF)      
      numFields = _stscanf(buf, executestmthdFormat, mode6);
#else
      numFields = oSscanf(buf, (SQLTCHAR*)_TEXT("EXECUTE"), mode6, NULL, NULL);
#endif     
      if (1 == numFields)
        {
        imode = atoi(mode6);
        }
      else 
        {
        imode = -1;
        goto exit;
        }
      
      if(imode < 0 || imode > 19)
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("the statement handle number is out of range (0~19) "));
        goto exit;
        }else
       clpstate->current_mhstmt = imode; //d4508 xh
      
      if(clpstate->mhstmt[imode] == 0) 
        
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT("No such statement handle exists"));
        goto exit;
        }
      else
        {  
        
        sqlrc=SQLExecute(clpstate->mhstmt[imode]);
        }
      
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      else
         {
        ErrorReport(clpstate);
        goto exit;
         }
      }
    // d4254qz ends
//#endif // end of nosscanf
    
    else if (0 == _tcsncmp(buf, _TEXT("$CHMOD "), 7))
      {
#if !defined(NOSSCANF) //08022001kmt
      numFields = _stscanf(buf, chmodFormat, mode);
      if (1 == numFields)
        {
        imode = atoi(mode);
        }
      else imode = -1;
#else //Palm CLP begin
      //parsing input to get correct value
      pstr = buf + 7;
      bstr = buf + _tcslen((SQLTCHAR*)buf);
      //remove spaces coming before the argument of $chmod
      while((*pstr == ' ') && pstr < bstr)
        pstr++;
      imode = _tcsAtoI((char*)pstr);         
      //imode = StrAToI(buf+7);
#endif  //Palm CLP end
      
      if (imode >= DB2eCLP_STMTATTR_UNCHANGED && 
        imode <= DB2eCLP_STMTATTR_SET_MODE9)
        {
        clpstate->stmtCtrl = imode;
        sqlrc = CLPSetStmtMode(clpstate);
        }
      }
    else if (0 == _tcsncmp(buf, _TEXT("IMPORT FROM "), 12))
      {
      sqlrc = DB2eImport(clpstate);
      }
    else if (0 == _tcsncmp(buf, _TEXT("EXPORT TO "), 10))
      {
      sqlrc = DB2eExport(clpstate);
      }
    else if (0 == _tcsncmp(buf, _TEXT("DESCRIBE "), 9))
      {
      CLPDescribeStatement(clpstate);
      }
    else if( 0 == _tcsncmp(buf, _TEXT("LIST TABLES"), 11))
      {
      writeOut(clpstate, (SQLTCHAR *)buf);
      _tcscpy(cmdstart, _TEXT("SELECT TNAME, NUMCOLS FROM \"DB2eSYSTABLES\""));
      sqlrc = CLPExecStatement(clpstate);
      }
    else if( 0 == _tcsncmp(buf, _TEXT("LIST COLUMNS"), 12))
      {
      writeOut(clpstate, (SQLTCHAR *)buf);
      _tcscpy(cmdstart, _TEXT("SELECT TNAME, CNAME, TYPE, KEYSEQ FROM \"DB2eSYSCOLUMNS\""));
      sqlrc = CLPExecStatement(clpstate);
      }
    else  if (0 == _tcsncmp(buf,_TEXT("$FILE"), 5 )) 
      {
#if !defined(oEPOC) && !defined(oWINCE) && !(defined oPALMOS)
      
      numFields = _stscanf(buf, fileFormat, inFile, outFile);
      
      if (2 == numFields)
        {
#if defined(oEPOC) // 04022001ctl
        _tcscpy(inPath, _TEXT("C:\\System\\Data\\DB2e\\"));
        _tcscpy(outPath, _TEXT("C:\\System\\Data\\DB2e\\"));
        _tcscat(inPath, inFile);
        _tcscat(outPath, outFile);
#else
        _tcscpy(inPath, inFile);
        _tcscpy(outPath, outFile);
#endif
        rc = processBatch(clpstate, inPath, outPath);
        }
      else if (1 == numFields)
        {
#if defined(oEPOC) // 04022001ctl
        _tcscpy(inPath,_TEXT( "C:\\System\\Data\\DB2e\\"));
        _tcscat(inPath, inFile);
#else
        _tcscpy(inPath, inFile);
#endif
        rc = processBatch(clpstate, inPath, 0);
        }
      else {
        print1 (clpstate, (SQLTCHAR*)_TEXT("Enter: $FILE [input file] [output file];\n"));
        }
#endif   
      }
    else if (0 == _tcsncmp(buf,_TEXT("$TABLE"), 6 )) // 04022001ctl
      {
      memset(outTable, 0, ID_LEN);
      memset(inTable, 0, ID_LEN);
      
#if defined(NOSSCANF) //Palm CLP //08022001kmt
      //Parsing the input to get correct table name and output filename
      pstr = buf + 6; // skip $TABLE
      // skip blanks
      numFields = 0;
      len = _tcslen((SQLTCHAR*)pstr);
      while (len > 0) {
        if (*pstr != ' ') {
          j = _tcslen((SQLTCHAR*)pstr);
          for (i = 0; i < j; i++) {
            if (numFields == 0) inTable[i] = oToUpper(*pstr);
            else if (numFields == 1) outTable[i] = oToUpper(*pstr);
            pstr++;
            len--;
            if (*pstr == ' ' || *pstr == '0') break;
            }
          numFields++;
          }
        else pstr++;
        len--;
        }
#else//Palm CLP end
      numFields = _stscanf(buf, tableFormat, inTable, outTable);
#endif
      
      if (numFields < 1 || numFields > 2)
        {
        print1 (clpstate, (SQLTCHAR*)_TEXT(CLP_TABLE_CMD_ERROR));
        goto exit;
        }
      
      tableCMDsql[0] = '\0';
      _tcscpy(tableCMDsql, _TEXT("SELECT ID, TXT FROM "));
      _tcscat(tableCMDsql, inTable);
      _tcscat(tableCMDsql, _TEXT(" ORDER BY ID"));
      
      // execute the SQL statements stored in the table
      rc = processTableCommands(clpstate, tableCMDsql, (numFields == 1 ? 0 : outTable));
      }
    else if (0 == _tcsncmp(buf, _TEXT("CONNECT TO "), 11)) // 03192001ctl
      {
      sqlrc = CLPReconnect(clpstate);
      }
    else if (0 == _tcsncmp(buf, _TEXT("AUTOCOMMIT OFF"), 14))
      {
      sqlrc = SQLSetConnectAttr(clpstate->hdbc, SQL_ATTR_AUTOCOMMIT, (void *)SQL_AUTOCOMMIT_OFF, 0 /* ignored */);
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      }
    else if (0 == _tcsncmp(buf, _TEXT("AUTOCOMMIT ON"), 13))
      {
      sqlrc = SQLSetConnectAttr(clpstate->hdbc, SQL_ATTR_AUTOCOMMIT, (void *)SQL_AUTOCOMMIT_ON, 0 /* ignored */);
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      }
    else if ((0 == _tcsncmp(buf, _TEXT("ROLLBACK WORK"), 13)) ||
      (0 == _tcsncmp(buf, _TEXT("ROLLBACK"), 8)))
      {
      sqlrc = SQLEndTran(SQL_HANDLE_DBC, clpstate->hdbc, SQL_ROLLBACK);
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      }
    else if ((0 == _tcsncmp(buf, _TEXT("COMMIT WORK"), 11)) ||
      (0 == _tcsncmp(buf, _TEXT("COMMIT"), 6)))
      {
      sqlrc = SQLEndTran(SQL_HANDLE_DBC, clpstate->hdbc, SQL_COMMIT);
      if (!sqlrc) writeOut(clpstate, (SQLTCHAR*)_TEXT(CLP_MSG_STMT_OK));
      }
    else if (0 == _tcsncmp(buf, _TEXT("VERSION"), 7))
      {
      sqlrc = SQLGetInfo(clpstate->hdbc, SQL_DBMS_VER, (void *) buf, BUFSIZE, &templen);
      if (!sqlrc) writeOut(clpstate, buf);
      }
    else if (0 == _tcsncmp(cmdstart,_TEXT("--"), 2))
      {
      }
    else
      {
      // Execute SQL statement otherwise
      sqlrc = CLPExecStatement(clpstate);
      }
    
exit:
      if (cmdbuf) { //08022001kmt
        free(cmdbuf);
        cmdbuf = 0;
        }
      // d3649 xh 06202001
#if defined(oEPOC)
      return sqlrc;
#else
      return CLP_CONTINUE;
#endif  
}
   

#ifdef oWINCE
/********************************************************************************
** o_Fread()
** This function reads a chuck of data from input stream into the buffer.
** Return number of elements read.
********************************************************************************/
OUINT32 o_Fread(void*   buffer,
                OUINT32 element_size,
                OUINT32 count,
                o_FILE*  stream)
{
    OUINT32 mb_nb_read = 0; //number of characters read (taking care of multibytes)
    OUINT32 nb_read = 0;
    OUINT32 sup_read = 0;
    char   *newbuf = 0;  
    char   *supbuf = 0;
    int     len = count + TEN_SPACES;
    OUINT32 newcount=0;
    int i;  
    SQLTCHAR *tmpbuf;

    // get some buffer
    newbuf = malloc(len);
    if (newbuf == 0) goto exit;
    memset(newbuf, 0, len);
    if (ReadFile(stream,
                 newbuf,
                 count,
                 &nb_read,
                 0) && nb_read > 0)
    {
        //checking for split multibytes characters      
        while((newcount<count)&&(mb_nb_read<nb_read)){

            newcount+=nTxCharSz((Uchar)newbuf[newcount]);
            mb_nb_read++;

            if (newcount>count) {
            
                len = newcount - count + TEN_SPACES;
                supbuf = malloc(len);
                if (supbuf == 0) goto exit;
                memset(supbuf, 0, len);

                ReadFile(stream,
                         supbuf,
                         newcount - count,
                         &sup_read,
                         0);
                                
                for(i=count;i<newcount;i++)
                    newbuf[i]=supbuf[i-count];              

                if(supbuf)
                    free(supbuf);
            }
        }        
        MultiByteToWideChar(CP_ACP, 0, newbuf, -1, buffer, newcount);           
    }
    else    
        mb_nb_read = 0; // If error then returns 0 byte read.  

    // free the local buffer
    free(newbuf);
  
    tmpbuf=buffer; //jca

    if ((tmpbuf[_tcslen(tmpbuf)-1] == '\n')&&(nb_read<count)) //jca
    {
        tmpbuf[_tcslen(tmpbuf)-1] = 0;
        mb_nb_read--;
    }


 exit:  
    return(mb_nb_read /*d4111kmt *element_size*/);

} /* o_Fread() */



/********************************************************************************
** o_Feof()
** This function returns EOF status of a stream.
********************************************************************************/
int o_Feof(o_FILE* stream)
{
    int status;

    /*--------------------------------------------------------------------------*/
    /*- Invokes system libraries.                                              -*/
    /*--------------------------------------------------------------------------*/
    if (ERROR_HANDLE_EOF == GetLastError())
        status = 1;
    else
        status = 0;

    return(status);
} /* o_Feof() */


/********************************************************************************
** o_Fflush()
** This function flushes the buffer.
********************************************************************************/
int o_Fflush(o_FILE* stream)
{
    return FlushFileBuffers(stream);
} /* o_Fflush() */


/********************************************************************************
** o_Fwrite()
** Write the buffer to the stream. 
********************************************************************************/
OUINT32 o_Fwrite(SQLTCHAR *buffer,
                 OUINT32   element_size,
                 OUINT32   count,
                 o_FILE    *stream)
{
    OUINT32 nb_write = 0;
    char   *newbuf = 0;
    int     len = element_size*count + TEN_SPACES;

    // character string in the buffer is in UCS-2 format, ie., 2 bytes/char
    // we need to convert to character using locale.
    newbuf = malloc(len);
    if (newbuf == 0) goto exit;
    memset(newbuf, 0, len);
    WideCharToMultiByte(CP_ACP,               // Converts from TCHAR to CHAR; codepage
                        0, 
                        (LPTSTR) buffer,
                        -1,                   // Null terminated string indicator
                        newbuf, 
                        len,                  // Buffer length
                        NULL,
                        NULL);
    len = strlen(newbuf);
    if (len <= 0) return(0);
    WriteFile(stream, newbuf, len, &nb_write, 0);

    // free the local buffer
    free(newbuf);

 exit:
    return(nb_write);

} /* o_Fwrite() */

/********************************************************************************
** o_Fclose()
** Close a file.
********************************************************************************/
int o_Fclose(o_FILE* stream)
{
    return CloseHandle(stream);
} /* o_Fclose() */

/********************************************************************************
** o_Fopen()
** Create or open a file.
********************************************************************************/
o_FILE *o_Fopen (SQLTCHAR *filename,
                 SQLTCHAR *mode)
{
    if (!_tcscmp(mode, _TEXT("rb")))
    {
        return CreateFile(filename,
                          GENERIC_READ,
                          FILE_SHARE_READ,
                          NULL,                 // OPEN_ALWAYS
                          OPEN_EXISTING,        // Requires that file should exist
                          FILE_ATTRIBUTE_NORMAL,
                          NULL);
    }
    else
    {
        return CreateFile(filename,
                          GENERIC_READ | GENERIC_WRITE,
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                          NULL,                 // OPEN_ALWAYS
                          CREATE_ALWAYS,        // If file not exist, then create it
                          FILE_ATTRIBUTE_NORMAL,
                          NULL);
    }
} /* o_Fopen() */

// return the uppercase of a given character s1
SQLTCHAR oToUpper(SQLTCHAR c) {
    if (c >= 'a' && c <= 'z') 
        return ((SQLTCHAR) (c - 32));
    else return c;
}

int nTxCharSz(const Uchar c) {
    unsigned char sz=1;  
    int NLS_Encoding = oGetCharEncoding(); 
   
    switch (NLS_Encoding) {
    case EncodingCP932:   
        if (FirstByteCP932(c)) sz = 2;
        break;
    case EncodingCP936:  
        if (FirstByteCP936(c)) sz = 2;
        break;
    case EncodingCP950:  
        if (FirstByteCP950(c)) sz = 2;
        break;
    case EncodingCP949:  
        if (FirstByteCP949(c)) sz = 2;
        break;
    case EncodingUTF8:   
    {           
        if (c & 0x80)
        {
            if (c & 0x20)
                sz = 3;
            else
                sz = 2;
        };
        break;
    };
    case EncodingCP1252:                
        break;
    default:            
        break;
    }
    return sz;
}

int oGetCharEncoding() {
    unsigned long encoding = EncodingUTF8;           

    if (-1 != OSS_Encoding) { 
        encoding = OSS_Encoding;
    }
    else
    {   
        int cp = GetACP();  
        
        if (cp)
            encoding = cp;    
    
        OSS_Encoding = encoding;
    };

    return encoding;
} 


#endif

