///////////////////////////////////////////////////////////////
//
// DB2 Everyplace Sample Application - CLP
// 
// This application demonstrates how to use CLI to
// write a simple command line processor that 
// accepts SQL statements and execute them.
// This application is portable to different platforms:
//    PalmOS, Neutrino, E. Linux, WinCE, Win32, and EPOC.
// It also includes import/export tools.
//
// File:  SampleCLPMain.c
//
///////////////////////////////////////////////////////////////


#include "SampleCLP.h"
#include "SampleCLPmsg.h"

#if defined(WIN32) && defined(UNICODE)
#include "wincon.h"  //d4357kmt
#endif

/********************************************************************************
** main()
**   This function simply loops and reads
**   from stdin.  When it finds the command
**   line delimiter ';', it calls 
**   CLPcommand_interpreter() to handle the
**   request.
********************************************************************************/
int main(int argc, char *argv[])
{
    SQLRETURN    rc;
    CLPState    *clpstate = 0;
    SQLTCHAR    *cmd = 0;
    SQLTCHAR    *pBuf;
    SQLTCHAR     c = 0;
    int          clp_status = CLP_CONTINUE;
    int          newCommand = 1;
    int          inString = 0;
    int          buflen = 0;
#define SZ     80          //d3753ctl
    SQLTCHAR     buffer[SZ]; //d3753ctl
    int          nb_read;
    SQLUINTEGER  idx;
    int i = 0; //d4254qz

    // Allocate CLPState structure
    clpstate = (CLPState *) malloc(sizeof(CLPState));
    if (!clpstate)
    {
        _tprintf ( _TEXT(CLP_INTERNAL_ERROR2) );
        exit(-1);
    }
    clpstate->mode  = CLP_INTERACT;
    clpstate->fpIN  = 0;
    clpstate->fpOUT = 0;

    //d4254qz 
	clpstate->mstmtno = 0;
	for(i = 0; i <20; i++)
    {
	   clpstate->mhstmt[i] = 0;
     clpstate->bindarray[i] = NULL;
    }

    // Start DB2 Everyplace engine
    rc = CLPStartEngine(clpstate);
    if (rc)
    {
        _tprintf ( _TEXT(CLP_DB2_NOSTART) );
        if (clpstate)
            free(clpstate);
        exit(-1);
    }
 
    if (SQL_SUCCESS == rc)
    {
        _tprintf( _TEXT(CLP_DB2_START) );
        pBuf = &(clpstate->cmd[0]);
        memset(pBuf, 0, MAX_CMD_LEN);
        _tprintf ( _TEXT(CLP_PROMPT) );
#if !defined(oPALMOS)
        o_Fflush(stdout);
#endif
      
        while ( !o_Feof(stdin) )
        {
            // read some bytes from the input file
            // process this line; we may hit end-of-line
#if !defined(oPALMOS)
  #if !defined(UNICODE) && !defined(_UNICODE)
            nb_read = o_Fread(buffer, 1, 1, stdin); //d3753ctl
  #else
    #ifdef WIN32
            if (!ReadConsole(GetStdHandle(STD_INPUT_HANDLE), buffer, 1, &nb_read, NULL)) //d4357kmt
              break;
    #endif
  #endif
#endif
            if (nb_read < 0)
                break;  // no more from the input file
            for (idx = 0; idx < nb_read; idx++) //d3753ctl
            {
                c = buffer[idx];

                if (EOF == c)
                    break;

                // check to see if it exceeds the maximum buffer length
                // reads MAX_CMD_LEN characters from stdin at at time only
                if (++buflen >= MAX_CMD_LEN)
                {
                    _tprintf ( _TEXT(CLP_INTERNAL_ERROR1) );
                    break;
                }
              
                if ( c != '\n' && newCommand )
                {
                    pBuf = &(clpstate->cmd[0]);
                    memset(pBuf, 0 , MAX_CMD_LEN);
                    newCommand = 0;       
                }
              
                // if quote
                if ( '\'' == c )
                { 
                    if (inString)
                        inString = 0;
                    else
                        inString = 1;
                    *pBuf++ = c;
                    continue;
                }
              
                // anything in between quotes
                if (inString)   
                {
                    if ('\n' == c ) {
                        _tprintf ( _TEXT(CLP_PROMPT) );
#if !defined(oPALMOS)
                        o_Fflush(stdout);
#endif    
                    }
                    else
                        *pBuf++ = c;
                    continue;
                }
              
                if ( ';' == c )
                {
                    if (inString)
                        *pBuf++ = c;
                    else  // command completes
                    {
                        *pBuf = 0;
                      
                        // handle command or SQL stmt
                        clp_status = CLPcommand_interpreter(clpstate);
                        if (CLP_QUIT == clp_status )
                            goto exit;
                        _tprintf ( _TEXT(CLP_PROMPT) );
#if !defined(oPALMOS)
                        o_Fflush(stdout);
#endif
                      
                        // reset states
                        newCommand = 1;
                        inString = 0;
                        buflen = 0;
                    }
                    continue;
                }
              
                // ignore carriage return CR
                if (0x0D == c || 0x0A == c) // CRLF
                {
                    continue;
                }
              
                *pBuf++ = c;
            }
        }
    }

 exit:
    // clean up
    if (clpstate)
    {
        CLPStopEngine(clpstate);
        free(clpstate);
    }

    return 0;
}

